"""
Export models from each Python module in this package.
"""
# pylint: disable=W0401

from .base import *
from .peer import *
from .staff import *
from .student_training import *
from .training import *



"""
Message step in the OpenAssessment XBlock.
"""

import datetime as dt

import pytz

from xblock.core import XBlock


class MessageMixin:
    """
    Message Mixin introduces all handlers for displaying the banner message

    MessageMixin is a Mixin for the OpenAssessmentBlock. Functions in the
    MessageMixin call into the OpenAssessmentBlock functions and will not work
    outside of OpenAssessmentBlock.
    """

    @XBlock.handler
    def render_message(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Render the message step.

        Args:
            data: Not used.

        Keyword Arguments:
            suffix: Not used.

        Returns:
            unicode: HTML content of the message banner.
        """
        # Retrieve the status of the workflow and information about deadlines.
        workflow = self.get_workflow_info()
        deadline_info = self._get_deadline_info()

        # Finds the cannonical status of the workflow and the is_closed status of the problem
        status = workflow.get('status')
        status_details = workflow.get('status_details', {})

        problem_is_closed = deadline_info.get('general').get('is_closed')

        # Finds the deadline info of the current step (defaults to submission)
        active_step_deadline_info = deadline_info.get(status, deadline_info.get("submission"))

        # Render the instruction message based on the status of the workflow
        # and the closed status.
        if self.teams_enabled and not self.valid_access_to_team_assessment():
            # If the learner is not on a team and hasn't submitted, warn them
            if status is None:
                path, context = self.render_message_no_team()
            else:
                path, context = 'openassessmentblock/message/oa_message_unavailable.html', {}
        elif status in ("done", "waiting"):
            path, context = self.render_message_complete(status_details)
        elif problem_is_closed or active_step_deadline_info.get('is_closed'):
            path, context = self.render_message_closed(active_step_deadline_info)
        elif status in ("self", "peer", "training"):
            path, context = self.render_message_incomplete(status, deadline_info)
        elif status == "cancelled":
            path, context = self.render_message_cancelled()
        elif status is None:
            path, context = self.render_message_open(deadline_info)
        else:
            # Default path leads to an "instruction-unavailable" block
            # Default context is empty
            path, context = 'openassessmentblock/message/oa_message_unavailable.html', {}

        context['xblock_id'] = self.get_xblock_id()
        return self.render_assessment(path, context)

    def render_message_incomplete(self, status, deadline_info):
        """
        Renders the "Incomplete" message state (User action still needed)

        Args:
            status (String): indicates the current step to be completed
            deadline_info (dict): The dictionary of boolean assessment near/closed states

        Returns:
            The path (String) and context (dict) to render the "Incomplete" message template
        """
        step_info = deadline_info.get(status, {})

        context = {
            status: True,
            f"{status}_approaching": step_info.get('approaching', False),
            f"{status}_not_released": (step_info.get("reason") == "start"),

            # Uses a static field in the XBlock to determine if the PeerAssessment Block
            # was able to pick up an assessment.
            "peer_not_available": self.no_peers,
        }

        return 'openassessmentblock/message/oa_message_incomplete.html', context

    def render_message_complete(self, status_details):
        """
        Renders the "Complete" message state (Either Waiting or Done)

        Args:
            status (String): indicates the canonical status of the workflow

        Returns:
            The path (String) and context (dict) to render the "Complete" message template
        """
        context = {
            "waiting": self.get_waiting_details(status_details),
        }

        return 'openassessmentblock/message/oa_message_complete.html', context

    def render_message_closed(self, status_info):
        """
        Renders the "Closed" message state

        Args:
            status_info (dict): The dictionary describing the closed status of the current step

        Returns:
            The path (String) and context (dict) to render the "Closed" template
        """

        reason = status_info.get("reason")

        context = {
            "not_yet_open": (reason == "start")
        }

        return 'openassessmentblock/message/oa_message_closed.html', context

    def render_message_open(self, deadline_info):
        """
        Renders the "Open" message state

        Args:
            deadline_info (dict): The dictionary of boolean assessment near/closed states

        Returns:
            The path (String) and context (dict) to render the "Open" template
        """

        submission_approaching = deadline_info.get("submission").get("approaching")

        context = {
            "approaching": submission_approaching
        }

        return 'openassessmentblock/message/oa_message_open.html', context

    def render_message_cancelled(self):
        """
        Renders the assessment "cancelled" message state

        Args:
            status (String): indicates the current step to be completed

        Returns:
            The path (String) and context (dict) to render the "cancelled" message template
        """
        context = {
            "is_team_assignment": self.is_team_assignment()
        }
        return 'openassessmentblock/message/oa_message_cancelled.html', context

    def _get_deadline_info(self):
        """
        Get detailed information about the standing of all deadlines.

        Args:
            None.

        Returns:
            dict with the following elements
                "submission" : dictionary on submission closure of this^ form
                "general" : dictionary on problem (all elements) closure of this^ form
                "peer": dictionary on peer closure of this^ form  *If Assessment has a Peer Section*
                "self": dictionary on self closure of this^ form  *If Assessment has a Self Section*

                this^ form:
                    "is_closed": (bool) Indicating whether or not that section has closed
                    "reason": (str) The reason that the section is closed (None if !is_closed)
                    "approaching": (bool) Indicates whether or not the section deadline is within a day.
        """

        # Methods which use datetime.deltatime to figure out if a deadline is approaching.
        now = dt.datetime.utcnow().replace(tzinfo=pytz.utc)

        def _is_approaching(date):
            # Determines if the deadline is within one day of now. (Approaching = True if so)
            delta = date - now
            return delta.days == 0

        # problem_info has form (is_closed, reason, start_date, due_date)
        problem_info = self.is_closed()

        # submission_info has form (is_closed, reason, start_date, due_date)
        submission_info = self.is_closed("submission")

        # The information we will always pass on to the user. Adds additional dicts on peer and self if applicable.
        deadline_info = {
            "submission": {
                "is_closed": submission_info[0],
                "reason": submission_info[1],
                "approaching": _is_approaching(submission_info[3])
            },
            "general": {
                "is_closed": problem_info[0],
                "reason": problem_info[1],
                "approaching": _is_approaching(problem_info[3])
            }
        }

        has_training = 'student-training' in self.assessment_steps

        if has_training:
            training_info = self.is_closed("student-training")
            training_dict = {
                "training": {
                    "is_closed": training_info[0],
                    "reason": training_info[1],
                    "approaching": _is_approaching(training_info[3])
                }
            }
            deadline_info.update(training_dict)

        has_peer = 'peer-assessment' in self.assessment_steps

        # peer_info has form (is_closed, reason, start_date, due_date)
        if has_peer:
            peer_info = self.is_closed("peer-assessment")
            peer_dict = {
                "peer": {
                    "is_closed": peer_info[0],
                    "reason": peer_info[1],
                    "approaching": _is_approaching(peer_info[3])
                }
            }
            deadline_info.update(peer_dict)

        has_self = 'self-assessment' in self.assessment_steps

        # self_info has form (is_closed, reason, start_date, due_date)
        if has_self:
            self_info = self.is_closed("self-assessment")
            self_dict = {
                "self": {
                    "is_closed": self_info[0],
                    "reason": self_info[1],
                    "approaching": _is_approaching(self_info[3])
                }
            }
            deadline_info.update(self_dict)

        return deadline_info

    def render_message_no_team(self):
        """
        Render the message when the user is not part of a team in the selected teamset.
        """
        if self.teamset_config:
            teamset_name = self.teamset_config.name
        else:
            teamset_name = '<id ' + self.selected_teamset_id + '>'
        context = {'teamset_name': teamset_name}
        return 'openassessmentblock/message/oa_message_no_team.html', context



"""
Validate changes to an XBlock before it is updated.
"""

from collections import Counter

from openassessment.assessment.api.student_training import validate_training_examples
from openassessment.assessment.serializers import InvalidRubric, rubric_from_dict
from openassessment.xblock.utils.data_conversion import convert_training_examples_list_to_dict
from openassessment.xblock.utils.resolve_dates import DateValidationError, InvalidDateFormat, resolve_dates


def _match_by_order(items, others):
    """
    Given two lists of dictionaries, each containing "order_num" keys,
    return a set of tuples, where the items in the tuple are dictionaries
    with the same "order_num" keys.

    Args:
        items (list of dict): Items to match, each of which must contain a "order_num" key.
        others (list of dict): Items to match, each of which must contain a "order_num" key.

    Returns:
        list of tuples, each containing two dictionaries

    Raises:
        IndexError: A dictionary does no contain a 'order_num' key.
    """
    # Sort each dictionary by its "name" key, then zip them and return
    def key_func(x):
        return x['order_num']
    return list(zip(sorted(items, key=key_func), sorted(others, key=key_func)))


def _duplicates(items):
    """
    Given an iterable of items, return a set of duplicate items in the list.

    Args:
        items (list): The list of items, which may contain duplicates.

    Returns:
        set: The set of duplicate items in the list.

    """
    counts = Counter(items)
    return {x for x in items if counts[x] > 1}


def _is_valid_assessment_sequence(assessments):
    """
    Check whether the sequence of assessments is valid. The rules enforced are:
        -must have one of staff-, peer-, or self- listed
        -in addition to those, only student-training is a valid entry
        -no duplicate entries
        -if staff-assessment is present, it must come last
        -if student-training is present, it must be followed at some point by peer-assessment

    Args:
        assessments (list of dict): List of assessment dictionaries.

    Returns:
        bool

    """
    sequence = [asmnt.get('name') for asmnt in assessments]
    required = ['staff-assessment', 'peer-assessment', 'self-assessment']
    optional = ['student-training']

    # at least one of required?
    if not any(name in required for name in sequence):
        return False

    # nothing except what appears in required or optional
    if any(name not in required + optional for name in sequence):
        return False

    # no duplicates
    if any(sequence.count(name) > 1 for name in sequence):
        return False

    # if using staff-assessment, it must come last
    if 'staff-assessment' in sequence and sequence[-1] != 'staff-assessment':
        return False

    # if using training, must be followed by peer at some point
    if 'student-training' in sequence:
        train_index = sequence.index('student-training')
        if 'peer-assessment' not in sequence[train_index:]:
            return False

    return True


def validate_assessments(assessments, current_assessments, is_released, _):
    """
    Check that the assessment dict is semantically valid. See _is_valid_assessment_sequence()
    above for a description of valid assessment sequences. In addition, enforces validation
    of several assessment-specific settings.

    If a question has been released, the type and number of assessment steps
    cannot be changed.

    Args:
        assessments (list of dict): list of serialized assessment models.
        current_assessments (list of dict): list of the current serialized
            assessment models. Used to determine if the assessment configuration
            has changed since the question had been released.
        is_released (boolean) : True if the question has been released.
        _ (function): The service function used to get the appropriate i18n text

    Returns:
        tuple (is_valid, msg) where
            is_valid is a boolean indicating whether the assessment is semantically valid
            and msg describes any validation errors found.
    """
    if not assessments:
        return False, _("This problem must include at least one assessment.")

    # Ensure that we support this sequence of assessments.
    if not _is_valid_assessment_sequence(assessments):
        msg = _("The assessment order you selected is invalid.")
        return False, msg

    for assessment_dict in assessments:
        # Number you need to grade is >= the number of people that need to grade you
        if assessment_dict.get('name') == 'peer-assessment':
            must_grade = assessment_dict.get('must_grade')
            must_be_graded_by = assessment_dict.get('must_be_graded_by')

            if must_grade is None or must_grade < 1:
                return False, _('In peer assessment, the "Must Grade" value must be a positive integer.')

            if must_be_graded_by is None or must_be_graded_by < 1:
                return False, _('In peer assessment, the "Graded By" value must be a positive integer.')

            if must_grade < must_be_graded_by:
                return False, _(
                    'In peer assessment, the "Must Grade" value must be greater than or equal to the "Graded By" value.'
                )

        # Student Training must have at least one example, and all
        # examples must have unique answers.
        if assessment_dict.get('name') == 'student-training':
            answers = []
            examples = assessment_dict.get('examples')
            if not examples:
                return False, _('You must provide at least one example response for learner training.')
            for example in examples:
                if example.get('answer') in answers:
                    return False, _('Each example response for learner training must be unique.')
                answers.append(example.get('answer'))

        # Staff grading must be required if it is the only step
        if assessment_dict.get('name') == 'staff-assessment' and len(assessments) == 1:
            required = assessment_dict.get('required')
            if not required:  # Captures both None and explicit False cases, both are invalid
                return False, _('The "required" value must be true if staff assessment is the only step.')

    if is_released:
        if len(assessments) != len(current_assessments):
            return False, _("The number of assessments cannot be changed after the problem has been released.")

        names = [assessment.get('name') for assessment in assessments]
        current_names = [assessment.get('name') for assessment in current_assessments]
        if names != current_names:
            return False, _("The assessment type cannot be changed after the problem has been released.")

    return True, ''


def validate_rubric(rubric_dict, current_rubric, is_released, _):
    """
    Check that the rubric is semantically valid.

    Args:
        rubric_dict (dict): Serialized Rubric model representing the updated state of the rubric.
        current_rubric (dict): Serialized Rubric model representing the current state of the rubric.
        is_released (bool): True if and only if the problem has been released.
        _ (function): The service function used to get the appropriate i18n text

    Returns:
        tuple (is_valid, msg) where
            is_valid is a boolean indicating whether the assessment is semantically valid
            and msg describes any validation errors found.
    """
    try:
        rubric_from_dict(rubric_dict)
    except InvalidRubric:
        return False, _('This rubric definition is not valid.')

    for criterion in rubric_dict['criteria']:
        # No duplicate option names within a criterion
        duplicates = _duplicates([option['name'] for option in criterion['options']])
        if duplicates:
            msg = _("Options in '{criterion}' have duplicate name(s): {duplicates}").format(
                criterion=criterion['name'], duplicates=", ".join(duplicates)
            )
            return False, msg

        # Some criteria may have no options, just written feedback.
        # In this case, written feedback must be required (not optional or disabled).
        if not criterion['options'] and criterion.get('feedback', 'disabled') != 'required':
            msg = _('Criteria with no options must require written feedback.')
            return False, msg

    # After a problem is released, authors are allowed to change text,
    # but nothing that would change the point value of a rubric.
    if is_released:

        # Number of prompts must be the same
        if len(rubric_dict['prompts']) != len(current_rubric['prompts']):
            return False, _('Prompts cannot be created or deleted after a problem is released.')

        # Number of criteria must be the same
        if len(rubric_dict['criteria']) != len(current_rubric['criteria']):
            return False, _('The number of criteria cannot be changed after a problem is released.')

        # Criteria names must be the same
        # We use criteria names as unique identifiers (unfortunately)
        # throughout the system.  Changing them mid-flight can cause
        # the grade page, for example, to raise 500 errors.
        # When we implement non-XML authoring, we might be able to fix this
        # the right way by assigning unique identifiers for criteria;
        # but for now, this is the safest way to avoid breaking problems
        # post-release.
        current_criterion_names = {criterion.get('name') for criterion in current_rubric['criteria']}
        new_criterion_names = {criterion.get('name') for criterion in rubric_dict['criteria']}
        if current_criterion_names != new_criterion_names:
            return False, _('Criteria names cannot be changed after a problem is released')

        # Number of options for each criterion must be the same
        for new_criterion, old_criterion in _match_by_order(rubric_dict['criteria'], current_rubric['criteria']):
            if len(new_criterion['options']) != len(old_criterion['options']):
                return False, _('The number of options cannot be changed after a problem is released.')

            else:
                for new_option, old_option in _match_by_order(new_criterion['options'], old_criterion['options']):
                    if new_option['points'] != old_option['points']:
                        return False, _('Point values cannot be changed after a problem is released.')

    return True, ''


def validate_dates(start, end, date_ranges, _):
    """
    Check that start and due dates are valid.

    Args:
        start (str): ISO-formatted date string indicating when the problem opens.
        end (str): ISO-formatted date string indicating when the problem closes.
        date_ranges (list of tuples): List of (start, end) pair for each submission / assessment.
        _ (function): The service function used to get the appropriate i18n text

    Returns:
        tuple (is_valid, msg) where
            is_valid is a boolean indicating whether the assessment is semantically valid
            and msg describes any validation errors found.
    """
    try:
        resolve_dates(start, end, date_ranges, _)
    except (DateValidationError, InvalidDateFormat) as ex:
        return False, str(ex)
    else:
        return True, ''


def validate_assessment_examples(rubric_dict, assessments, _):
    """
    Validate assessment training examples.

    Args:
        rubric_dict (dict): The serialized rubric model.
        assessments (list of dict): List of assessment dictionaries.
        _ (function): The service function used to get the appropriate i18n text

    Returns:
        tuple (is_valid, msg) where
            is_valid is a boolean indicating whether the assessment is semantically valid
            and msg describes any validation errors found.

    """
    for asmnt in assessments:
        if asmnt['name'] == 'student-training':

            examples = convert_training_examples_list_to_dict(asmnt['examples'])

            # Must have at least one training example
            if not examples:
                return False, _(
                    "Learner training must have at least one training example."
                )

            # Delegate to the student training API to validate the
            # examples against the rubric.
            errors = validate_training_examples(rubric_dict, examples)
            if errors:
                return False, "; ".join(errors)

    return True, ''


def validator(oa_block, _, strict_post_release=True):
    """
    Return a validator function configured for the XBlock.
    This will validate assessments, rubrics, and dates.

    Args:
        oa_block (OpenAssessmentBlock): The XBlock being updated.
        _ (function): The service function used to get the appropriate i18n text

    Keyword Arguments:
        strict_post_release (bool): If true, restrict what authors can update once
            a problem has been released.

    Returns:
        callable, of a form that can be passed to `update_from_xml`.
    """
    # Import is placed here to avoid model import at project startup.
    from submissions.api import MAX_TOP_SUBMISSIONS

    def _inner(rubric_dict, assessments, leaderboard_show=0, submission_start=None, submission_due=None):
        """ Validator method. """

        is_released = strict_post_release and oa_block.is_released()

        # Assessments
        current_assessments = oa_block.rubric_assessments
        success, msg = validate_assessments(assessments, current_assessments, is_released, _)
        if not success:
            return False, msg

        # Rubric
        current_rubric = {
            'prompts': oa_block.prompts,
            'criteria': oa_block.rubric_criteria
        }
        success, msg = validate_rubric(rubric_dict, current_rubric, is_released, _)
        if not success:
            return False, msg

        # Training examples
        success, msg = validate_assessment_examples(rubric_dict, assessments, _)
        if not success:
            return False, msg

        # Dates
        submission_dates = [(submission_start, submission_due)]
        assessment_dates = [(asmnt.get('start'), asmnt.get('due')) for asmnt in assessments]
        success, msg = validate_dates(oa_block.start, oa_block.due, submission_dates + assessment_dates, _)
        if not success:
            return False, msg

        # Leaderboard
        if leaderboard_show < 0 or leaderboard_show > MAX_TOP_SUBMISSIONS:
            return False, _("Leaderboard number is invalid.")

        # Success!
        return True, ''

    return _inner


def validate_submission(submission, prompts, _, text_response='required'):
    """
    Validate submission dict.

    Args:
        submission (list of unicode): Responses for the prompts.
        prompts (list of dict): The prompts from the problem definition.
        _ (function): The service function used to get the appropriate i18n text.

    Returns:
        tuple (is_valid, msg) where
            is_valid is a boolean indicating whether the submission is semantically valid
            and msg describes any validation errors found.
    """

    message = _("The submission format is invalid.")

    if not isinstance(submission, list):
        return False, message

    if text_response == 'required' and len(submission) != len(prompts):
        return False, message

    for submission_part in submission:
        if not isinstance(submission_part, str):
            return False, message

    return True, ''



"""Collection of JSON handlers for legacy view"""
import logging

from openassessment.assessment.errors import (
    PeerAssessmentInternalError,
    PeerAssessmentRequestError,
    PeerAssessmentWorkflowError,
)
from openassessment.workflow.errors import AssessmentWorkflowError

from openassessment.xblock.utils.data_conversion import (
    clean_criterion_feedback,
    create_rubric_dict,
)

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

messages = {
    "must_submit": "You must submit a response before you can perform a peer assessment.",
    "already_submitted": "This feedback has already been submitted or the submission has been cancelled.",
    "could_not_submit": "Your peer assessment could not be submitted.",
    "could_not_update": "Could not update workflow status.",
    "could_not_load": "Could not load peer assessment.",
}


def create_peer_assessment(api_data, data):
    # Import is placed here to avoid model import at project startup.
    from openassessment.assessment.api import peer as peer_api

    config_data = api_data.config_data
    submission_uuid = api_data.workflow_data.submission_uuid
    peer_assessment_data = api_data.peer_assessment_data()

    # Create the assessment
    assessment = peer_api.create_assessment(
        submission_uuid,
        peer_assessment_data.student_item["student_id"],
        data["options_selected"],
        clean_criterion_feedback(
            config_data.rubric_criteria_with_labels,
            data["criterion_feedback"],
        ),
        data["overall_feedback"],
        create_rubric_dict(
            config_data.prompts,
            config_data.rubric_criteria_with_labels,
        ),
        peer_assessment_data.assessment["must_be_graded_by"],
    )

    return assessment


def peer_assess(api_data, data, suffix=""):  # pylint: disable=unused-argument
    """Place a peer assessment into OpenAssessment system

    Assess a Peer Submission.  Performs basic workflow validation to ensure
    that an assessment can be performed as this time.

    Args:
        data (dict): A dictionary containing information required to create
            a new peer assessment.  This dict should have the following attributes:
            `submission_uuid` (string): The unique identifier for the submission being assessed.
            `options_selected` (dict): Dictionary mapping criterion names to option values.
            `overall_feedback` (unicode): Written feedback for the submission as a whole.
            `criterion_feedback` (unicode): Written feedback per the criteria for the submission.

    Returns:
        Dict with keys "success" (bool) indicating success/failure.
        and "msg" (unicode) containing additional information if an error occurs.

    """
    translate = api_data.config_data.translate
    submission_uuid = api_data.workflow_data.submission_uuid

    def failure_response(reason):
        return {"success": False, "msg": translate(reason)}

    if submission_uuid is None:
        return failure_response(messages["must_submit"])

    step_data = api_data.peer_assessment_data()
    submission = step_data.get_peer_submission() or {}
    uuid_server = submission.get("uuid", None)
    uuid_client = data.get("submission_uuid", None)

    if uuid_server != uuid_client:
        logger.warning(
            "Irrelevant assessment submission: expected '%s', got '%s'",
            uuid_server,
            uuid_client,
        )
        return failure_response(messages["already_submitted"])

    if step_data.assessment:
        try:
            assessment = create_peer_assessment(api_data, data)
            # Emit analytics event...
            api_data.config_data.publish_assessment_event("openassessmentblock.peer_assess", assessment)
        except (PeerAssessmentRequestError, PeerAssessmentWorkflowError):
            logger.warning(
                "Peer API error for submission UUID %s",
                submission_uuid,
                exc_info=True,
            )
            return failure_response(messages["could_not_submit"])
        except PeerAssessmentInternalError:
            logger.exception(
                "Peer API internal error for submission UUID: %s",
                submission_uuid,
            )
            return failure_response(messages["could_not_submit"])

        # Update both the workflow that the submission we"re assessing
        # belongs to, as well as our own (e.g. have we evaluated enough?)
        try:
            update = api_data.workflow_data.update_workflow_status
            if assessment:
                update(submission_uuid=assessment["submission_uuid"])
            update()
        except AssessmentWorkflowError:
            logger.exception(
                "Workflow error occurred when submitting peer assessment for submission %s",
                submission_uuid,
            )
            return failure_response(messages["could_not_update"])

        # Temp kludge until we fix JSON serialization for datetime
        assessment["scored_at"] = str(assessment["scored_at"])

        return {"success": True, "msg": ""}

    return failure_response(messages["could_not_load"])



""" Django backend for file upload. """


import os
from urllib.parse import urljoin

from django.conf import settings
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.urls import reverse

from .base import BaseBackend


class Backend(BaseBackend):
    """
    Manage openassessment student files uploaded using the default django storage settings.
    """
    def get_upload_url(self, key, content_type):
        """
        Return the URL pointing to the ORA2 django storage upload endpoint.
        """
        return reverse("openassessment-django-storage", kwargs={'key': key})

    def get_download_url(self, key):
        """
        Return the django storage download URL for the given key.

        Returns None if no file exists at that location.
        """
        path = self._get_file_path(key)
        if default_storage.exists(path):
            storage_path = default_storage.url(path)

            # Return a fully-qualified URL
            lms_url = getattr(settings, 'LMS_ROOT_URL', '')
            return urljoin(lms_url, storage_path)
        return None

    def upload_file(self, key, content):
        """
        Upload the given file content to the keyed location.
        """
        path = self._get_file_path(key)
        saved_path = default_storage.save(path, ContentFile(content))
        return saved_path

    def remove_file(self, key):
        """
        Remove the file at the given keyed location.

        Returns True if the file exists, and was removed.
        Returns False if the file does not exist, and so was not removed.
        """
        path = self._get_file_path(key)
        if default_storage.exists(path):
            default_storage.delete(path)
            return True
        return False

    def _get_file_name(self, key):
        """
        Returns the name of the keyed file.

        Since the backend storage may be folders, or it may use pseudo-folders,
        make sure the filename doesn't include any path separators.
        """
        file_name = key.replace("..", "").strip("/ ")
        file_name = file_name.replace(os.sep, "_")
        return file_name

    def _get_file_path(self, key):
        """
        Returns the path to the keyed file, including the storage prefix.
        """
        path = self._get_key_name(self._get_file_name(key))
        return path



"""
Exposed api for ORA XBlock workflows.
"""


class WorkflowAPI:
    def __init__(self, block):
        self._block = block

    def get_workflow_info(self, submission_uuid=None):
        return self._block.get_workflow_info(submission_uuid)

    @property
    def workflow(self):
        return self.get_workflow_info()

    @property
    def has_workflow(self):
        return bool(self.workflow)

    @property
    def has_status(self):
        return bool(self.status)

    @property
    def status_details(self):
        return self.workflow.get("status_details", {})

    @property
    def is_peer_complete(self):
        return self.status_details.get("peer", {}).get("complete", False)

    @property
    def is_peer_skipped(self):
        return self.status_details.get("peer", {}).get("skipped", False)

    @property
    def is_self_complete(self):
        return self.status_details.get("self", {}).get("complete", False)

    @property
    def is_cancelled(self):
        return self.status == "cancelled"

    @property
    def is_done(self):
        return self.status == "done"

    @property
    def is_waiting(self):
        return self.status == "waiting"

    @property
    def is_self(self):
        return self.status == "self"

    @property
    def is_training(self):
        return self.status == "training"

    @property
    def is_peer(self):
        return self.status == "peer"

    @property
    def submission_uuid(self):
        return self._block.submission_uuid

    @property
    def workflow_requirements(self):
        return self._block.workflow_requirements

    @property
    def status(self):
        return self.workflow.get("status")

    def get_workflow_status_counts(self):
        return self._block.get_workflow_status_counts()

    def get_workflow_cancellation_info(self, submission_uuid):
        return self._block.get_workflow_cancellation_info(submission_uuid)

    def get_course_workflow_settings(self):
        return self._block.get_course_workflow_settings()

    def update_workflow_status(self, submission_uuid=None):
        self._block.update_workflow_status(submission_uuid)

    def create_workflow(self, submission_uuid):
        self._block.create_workflow(submission_uuid)

    def create_team_workflow(self, submission_uuid):
        self._block.create_team_workflow(submission_uuid)

    def get_team_workflow_info(self, team_submission_uuid=None):
        return self._block.get_team_workflow_info(team_submission_uuid)

    def get_team_submission_uuid(self):
        return self._block.get_team_submission_uuid()

    def get_team_workflow_status_counts(self):
        return self._block.get_team_workflow_status_counts()

    def get_team_workflow_cancellation_info(self, team_submission_uuid):
        return self._block.get_team_workflow_cancellation_info(team_submission_uuid)



"""
The Peer Assessment view Mixin
"""
from webob import Response
from openassessment.xblock.utils.defaults import DEFAULT_RUBRIC_FEEDBACK_TEXT

from openassessment.xblock.utils.user_data import get_user_preferences


template_paths = {
    "unavailable": "openassessmentblock/peer/oa_peer_unavailable.html",
    "cancelled": "openassessmentblock/peer/oa_peer_cancelled.html",
    "complete": "openassessmentblock/peer/oa_peer_complete.html",
    "turbo_mode": "openassessmentblock/peer/oa_peer_turbo_mode.html",
    "turbo_mode_waiting": "openassessmentblock/peer/oa_peer_turbo_mode_waiting.html",
    "closed": "openassessmentblock/peer/oa_peer_closed.html",
    "assessment": "openassessmentblock/peer/oa_peer_assessment.html",
    "waiting": "openassessmentblock/peer/oa_peer_waiting.html",
}


def peer_context(config_data, step_data, peer_sub=None):
    user_preferences = get_user_preferences(config_data.user_service)
    context_dict = {
        "rubric_criteria": config_data.rubric_criteria_with_labels,
        "allow_multiple_files": config_data.allow_multiple_files,
        "allow_latex": config_data.allow_latex,
        "prompts_type": config_data.prompts_type,
        "user_timezone": user_preferences["user_timezone"],
        "user_language": user_preferences["user_language"],
        "xblock_id": config_data.get_xblock_id(),
    }

    if config_data.rubric_feedback_prompt is not None:
        context_dict["rubric_feedback_prompt"] = config_data.rubric_feedback_prompt

    if config_data.rubric_feedback_default_text is not None:
        context_dict["rubric_feedback_default_text"] = config_data.rubric_feedback_default_text

    # We display the due date whether the problem is open or closed.
    # If no date is set, it defaults to the distant future, in which
    # case we don"t display the date.
    if step_data.is_due:
        context_dict["peer_due"] = step_data.due_date

    assessment = step_data.assessment
    translate = config_data.translate

    if assessment:
        count = step_data.has_finished[1]

        if step_data.continue_grading:
            context_dict["submit_button_text"] = translate("Submit your assessment and review another response")
        elif assessment["must_grade"] - count == 1:
            context_dict["submit_button_text"] = translate("Submit your assessment and move to next step")
        else:
            context_dict["submit_button_text"] = translate(
                "Submit your assessment and move to response #{response_number}"
            ).format(response_number=(count + 2))

        context_dict.update(
            {
                "must_grade": assessment["must_grade"],
                "graded": count,
                "review_num": count + 1,
            }
        )

    if step_data.is_not_available_yet:
        context_dict["peer_start"] = step_data.start_date

    if peer_sub:
        context_dict.update(
            {
                "peer_submission": step_data.get_submission_dict(peer_sub),
                # Determine if file upload is supported for this XBlock.
                "file_upload_type": step_data.file_upload_type,
                "peer_file_urls": step_data.get_download_urls(peer_sub),
            }
        )

    return context_dict


def peer_path_and_context(api_data, continue_grading):
    """
    Return the template path and context for rendering the peer assessment step.

    Args:
        continue_grading (bool): If true, the user has chosen to continue grading.

    Returns:
        tuple of (template_path, context_dict)

    """
    step_data = api_data.peer_assessment_data(continue_grading)

    def path_and_context(path_key, peer_sub=None):
        return (
            template_paths[path_key],
            peer_context(api_data.config_data, step_data, peer_sub),
        )

    if step_data.is_cancelled:
        return path_and_context("cancelled")

    # Once a student has completed a problem, it stays complete,
    # so this condition needs to be first.
    if (step_data.is_complete) and not step_data.continue_grading:
        return path_and_context("complete")

    # Allow continued grading even if the problem due date has passed
    if step_data.continue_grading and step_data.student_item:
        peer_sub = step_data.get_peer_submission()
        if peer_sub:
            return path_and_context("turbo_mode", peer_sub)
        else:
            return path_and_context("turbo_mode_waiting")

    if step_data.is_past_due:
        return path_and_context("closed")

    if step_data.is_not_available_yet:
        return path_and_context("unavailable")

    if step_data.is_peer or step_data.is_skipped:
        peer_sub = step_data.get_peer_submission()
        if peer_sub:
            return path_and_context("assessment", peer_sub)
        else:
            return path_and_context("waiting")

    return path_and_context("unavailable")


def render_peer_assessment(api_data, continue_grading):
    """Renders the Peer Assessment HTML section of the XBlock

    Generates the peer assessment HTML for the first section of an Open
    Assessment XBlock. See OpenAssessmentBlock.render_assessment() for
    more information on rendering XBlock sections.

    Args:
        data (dict): May contain an attribute "continue_grading", which
            allows a student to continue grading peers past the required
            number of assessments.

    """
    if "peer-assessment" not in api_data.config_data.assessment_steps:
        return Response("")

    path, context_dict = peer_path_and_context(api_data, continue_grading)

    # For backwards compatibility, if no feedback default text has been
    # set, use the default text
    if "rubric_feedback_default_text" not in context_dict:
        context_dict["rubric_feedback_default_text"] = DEFAULT_RUBRIC_FEEDBACK_TEXT

    return api_data.config_data.render_assessment(path, context_dict)



# pylint: disable=import-error
"""
A module for containing various inline runtime class imports
Functions in this module should take no args and simply import and return the <class> object.
"""


def import_block_structure_transformers():
    """
    Helper method that imports BlockStructureTransformers from edx platform at runtime.
    """
    from openedx.core.djangoapps.content.block_structure.transformers import BlockStructureTransformers
    return BlockStructureTransformers


def import_course_waffle_flag():
    """
    Helper method that imports CourseWaffleFlag from edx-platform at runtime.
    https://github.com/openedx/edx-platform/blob/master/openedx/core/djangoapps/waffle_utils/__init__.py#L345
    """
    from openedx.core.djangoapps.waffle_utils import CourseWaffleFlag
    return CourseWaffleFlag


def import_external_id():
    """
    Helper method that imports ExternalId from edx platform at runtime.
    """
    from openedx.core.djangoapps.external_user_ids.models import ExternalId
    return ExternalId


def import_waffle_flag():
    """
    Helper method that imports WaffleFlag from edx_toggles at runtime.
    """
    from edx_toggles.toggles import WaffleFlag
    return WaffleFlag



"""
Data layer for ORA

XBlock handlers which surface info about an ORA, instead of being tied to views.
"""
from xblock.core import XBlock

from openassessment.xblock.ui_mixins.mfe.serializers import OraBlockInfoSerializer


class MfeMixin:
    @XBlock.json_handler
    def get_block_info(self, data, suffix=""):  # pylint: disable=unused-argument
        block_info = OraBlockInfoSerializer(self)
        return block_info.data



"""
Views for the old-style ORA UI
"""

from xblock.core import XBlock

from .peer_assessments.views import render_peer_assessment, peer_path_and_context
from .self_assessments.views import render_self_assessment, self_path_and_context
from .staff_assessments.views import render_staff_assessment, staff_path_and_context
from .student_training.views import render_student_training, training_path_and_context
from .submissions.views import render_submission, get_submission_path, get_submission_context


class LegacyViewsMixin:
    @XBlock.handler
    def render_submission(self, data, suffix=""):  # pylint: disable=unused-argument
        return render_submission(self.config_data, self.submission_data)

    # NOTE - Temporary surfacing for testing / refactoring
    def submission_path_and_context(self):
        return get_submission_path(self.submission_data), get_submission_context(self.config_data, self.submission_data)

    @XBlock.handler
    def render_peer_assessment(self, data, suffix=""):  # pylint: disable=unused-argument
        continue_grading = data.params.get("continue_grading", False)
        peer_assessment_data = self.peer_assessment_data(continue_grading)
        if peer_assessment_data.is_cancelled:
            self.no_peers = True
        if peer_assessment_data.is_peer or peer_assessment_data.is_skipped:
            if peer_assessment_data.get_peer_submission():
                self.no_peers = False
            else:
                self.no_peers = True
        return render_peer_assessment(self.api_data, continue_grading)

    # NOTE - Temporary surfacing for testing / refactoring
    def peer_path_and_context(self, continue_grading=False):
        return peer_path_and_context(self.api_data, continue_grading)

    # NOTE - Temporary surfacing for testing / refactoring
    def self_path_and_context(self):
        return self_path_and_context(self.api_data)

    # NOTE - Temporary surfacing for testing / refactoring
    def staff_path_and_context(self):
        return staff_path_and_context(self.api_data)

    # NOTE - Temporary surfacing for testing / refactoring
    def training_path_and_context(self):
        return training_path_and_context(self.api_data)

    @XBlock.handler
    def render_self_assessment(self, data, suffix=""):  # pylint: disable=unused-argument
        step_data = self.api_data.self_assessment_data
        if step_data.is_cancelled:
            self.no_peers = True
        return render_self_assessment(self.api_data)

    @XBlock.handler
    def render_staff_assessment(self, data, suffix=""):  # pylint: disable=unused-argument
        return render_staff_assessment(self.api_data)

    @XBlock.handler
    def render_student_training(self, data, suffix=""):  # pylint: disable=unused-argument
        return render_student_training(self.api_data)



"""
Public interface for staff grading of team assignments, used by students/course staff.
"""


import logging

from django.db import DatabaseError
from django.utils.timezone import now

from submissions import team_api as team_submissions_api

from openassessment.assessment.api.staff import _complete_assessment
from openassessment.assessment.errors import StaffAssessmentInternalError, StaffAssessmentRequestError
from openassessment.assessment.models import Assessment, TeamStaffWorkflow, InvalidRubricSelection
from openassessment.assessment.serializers import InvalidRubric, full_assessment_dict
from openassessment.assessment.score_type_constants import STAFF_TYPE

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def submitter_is_finished(team_submission_uuid, team_requirements):  # pylint: disable=unused-argument
    """
    Determine if the submitter has finished their requirements for staff
    assessment. Always returns True.

    Args:
        team_submission_uuid (str): Not used.
        team_requirements (dict): Not used.

    Returns:
        True

    """
    return True


def assessment_is_finished(team_submission_uuid, staff_requirements, _):
    """
    Determine if the staff assessment step of the given team submission is completed.
    This checks to see if staff have completed the assessment.

    Args:
        team_submission_uuid (str): The UUID of the submission being graded.
        staff_requirements (dict): Any variables that may effect this state.

    Returns:
        True if a staff assessment has been completed for this team submission or if not required.
    """
    # Requirements of None means we can't make any assumptions about the done-ness of this step
    if staff_requirements is None:
        return False

    if staff_requirements.get('required', False):
        return bool(get_latest_staff_assessment(team_submission_uuid))

    return True


def on_init(team_submission_uuid):
    """
    Create a new team staff workflow for a student item and submission.

    Creates a unique team staff workflow for a student item, associated with a
    team submission.

    Note that the staff workflow begins things in on_init() instead of
    on_start(), because staff shoud be able to access the submission
    regardless of which state the workflow is currently in.

    Args:
        team_submission_uuid (str): The team submission associated with this workflow.

    Returns:
        None

    Raises:
        StaffAssessmentInternalError: Raised when there is an internal error
            creating the Workflow.
    """
    try:
        team_submission = team_submissions_api.get_team_submission(team_submission_uuid)
        TeamStaffWorkflow.objects.get_or_create(
            course_id=team_submission['course_id'],
            item_id=team_submission['item_id'],
            team_submission_uuid=team_submission_uuid,
            # submission_uuid is currently not used in any logic in TeamStaffWorkflow, so we don't
            # realy care which submission is chosen and it doesn't need to match the TeamAssessment Workflow.
            # It must be filled because of the unique constraint on the field (can't have multiple '' values)
            submission_uuid=team_submission['submission_uuids'][0],
        )
    except DatabaseError as ex:
        error_message = (
            "An internal error occurred while creating a new team staff workflow for team submission {}"
        ).format(team_submission_uuid)
        logger.exception(error_message)
        raise StaffAssessmentInternalError(error_message) from ex


def on_cancel(team_submission_uuid):
    """
    Cancel the team staff workflow for submission.

    Sets the cancelled_at field in team staff workflow.

    Args:
        team_submission_uuid (str): The team_submission UUID associated with this workflow.

    Returns:
        None

    """
    try:
        workflow = TeamStaffWorkflow.objects.get(team_submission_uuid=team_submission_uuid)
        workflow.cancelled_at = now()
        workflow.save(update_fields=['cancelled_at'])
    except TeamStaffWorkflow.DoesNotExist:
        # If we can't find a workflow, then we don't have to do anything to
        # cancel it.
        pass
    except DatabaseError as ex:
        error_message = (
            "An internal error occurred while cancelling the team staff workflow for submission {}"
        ).format(team_submission_uuid)
        logger.exception(error_message)
        raise StaffAssessmentInternalError(error_message) from ex


def get_score(team_submission_uuid, staff_requirements, course_settings):  # pylint: disable=unused-argument
    """
    Generate a score based on a completed assessment for the given team submission.
    If no assessment has been completed for this submission, this will return
    None.

    Args:
        team_submission_uuid (str): The UUID for the submission to get a score for.
        staff_requirements (dict): Not used.
        course_settings (dict): Not used.

    Returns:
        A dictionary with the points earned, points possible,
        contributing_assessments, and staff_id information.

    """
    assessment = get_latest_staff_assessment(team_submission_uuid)
    if not assessment:
        return None

    return {
        "points_earned": assessment["points_earned"],
        "points_possible": assessment["points_possible"],
        "contributing_assessments": [assessment["id"]],
        "staff_id": assessment["scorer_id"],
    }


def get_latest_staff_assessment(team_submission_uuid):
    """
    Retrieve the latest staff assessment for a team submission.

    Args:
        team_submission_uuid (str): The UUID of the team submission being assessed.

    Returns:
        dict: The serialized assessment model
        or None if no assessments are available

    Raises:
        StaffAssessmentInternalError if there are problems connecting to the database.

    Example usage:

    >>> get_latest_staff_assessment('10df7db776686822e501b05f452dc1e4b9141fe5')
    {
        'points_earned': 6,
        'points_possible': 12,
        'scored_at': datetime.datetime(2014, 1, 29, 17, 14, 52, 649284 tzinfo=<UTC>),
        'scorer': "staff",
        'feedback': ''
    }

    """
    try:
        # Get the reference submissions
        team_submission = team_submissions_api.get_team_submission(team_submission_uuid)

        # Return the most-recently graded assessment for any team member's submisison
        assessment = Assessment.objects.filter(
            submission_uuid__in=team_submission['submission_uuids'],
            score_type=STAFF_TYPE,
        ).first()
    except DatabaseError as ex:
        msg = (
            "An error occurred while retrieving staff assessments "
            "for the submission with UUID {uuid}: {ex}"
        ).format(uuid=team_submission_uuid, ex=ex)
        logger.exception(msg)
        raise StaffAssessmentInternalError(msg) from ex

    if assessment:
        return full_assessment_dict(assessment)

    return None


def get_assessment_scores_by_criteria(team_submission_uuid):
    """Get the staff score for each rubric criterion

    Args:
        team_submission_uuid (str): The team submission uuid is used to get the
            assessment used to score this submission.

    Returns:
        (dict): A dictionary of rubric criterion names, with a score of
            the staff assessments.

    Raises:
        StaffAssessmentInternalError: If any error occurs while retrieving
            information from the scores, an error is raised.
    """
    try:
        # Get most recently graded assessment for a team submission
        team_submission = team_submissions_api.get_team_submission(team_submission_uuid)
        assessments = list(
            Assessment.objects.filter(
                submission_uuid__in=team_submission['submission_uuids'],
                score_type=STAFF_TYPE,
            )[:1]
        )

        scores = Assessment.scores_by_criterion(assessments)
        # Since this is only being sent one score, the median score will be the
        # same as the only score.
        return Assessment.get_median_score_dict(scores)
    except DatabaseError as ex:
        error_message = f"Error getting staff assessment scores for {team_submission_uuid}"
        logger.exception(error_message)
        raise StaffAssessmentInternalError(error_message) from ex


def get_submission_to_assess(course_id, item_id, scorer_id):
    """
    Get a team submission for staff evaluation.

    Retrieves a team submission for assessment for the given staff member.

    Args:
        course_id (str): The course that we would like to fetch submissions from.
        item_id (str): The student_item (problem) that we would like to retrieve submissions for.
        scorer_id (str): The user id of the staff member scoring this submission

    Returns:
        dict: A student submission for assessment. This contains a 'student_item',
            'attempt_number', 'submitted_at', 'created_at', and 'answer' field to be
            used for assessment.

    Raises:
        StaffAssessmentInternalError: Raised when there is an internal error
            retrieving staff workflow information.

    Examples:
        >>> get_submission_to_assess("a_course_id", "an_item_id", "a_scorer_id")
        {
            'student_item': 2,
            'attempt_number': 1,
            'submitted_at': datetime.datetime(2014, 1, 29, 23, 14, 52, 649284, tzinfo=<UTC>),
            'created_at': datetime.datetime(2014, 1, 29, 17, 14, 52, 668850, tzinfo=<UTC>),
            'answer': { ... }
        }

    """
    team_submission_uuid = TeamStaffWorkflow.get_submission_for_review(course_id, item_id, scorer_id)
    if team_submission_uuid:
        try:
            submission_data = team_submissions_api.get_team_submission(team_submission_uuid)
            return submission_data
        except DatabaseError as ex:
            error_message = (
                "Could not find a team submission with the uuid {}"
            ).format(team_submission_uuid)
            logger.exception(error_message)
            raise StaffAssessmentInternalError(error_message) from ex
    else:
        logger.info("No team submission found for staff to assess (%s, %s)", course_id, item_id)
        return None


def get_staff_grading_statistics(course_id, item_id):
    """
    Returns the number of graded, ungraded, and in-progress team submissions for staff grading.

    Args:
        course_id (str): The course that this problem belongs to
        item_id (str): The student_item (problem) that we want to know statistics about.

    Returns:
        dict: a dictionary that contains the following keys: 'graded', 'ungraded', and 'in-progress'
    """
    return TeamStaffWorkflow.get_workflow_statistics(course_id, item_id)


def create_assessment(
        team_submission_uuid,
        scorer_id,
        options_selected,
        criterion_feedback,
        overall_feedback,
        rubric_dict,
        scored_at=None
):
    """
    Creates an assessment for each member of the submitting team.

    Closely mirrors openassessment.assessment.api.staff.py::create_assessment

    Can use _complete_assessment from Staff API as is, but has the side-effect
    of only associating the last graded assessment with the workflow

    Returns:
        dict: the Assessment model, serialized as a dict.
    """
    try:
        try:
            scorer_workflow = TeamStaffWorkflow.objects.get(team_submission_uuid=team_submission_uuid)
        except TeamStaffWorkflow.DoesNotExist:
            scorer_workflow = None

        # Get the submissions for a team
        team_submission = team_submissions_api.get_team_submission(team_submission_uuid)

        assessment_dicts = []
        for submission_uuid in team_submission['submission_uuids']:
            assessment = _complete_assessment(
                submission_uuid,
                scorer_id,
                options_selected,
                criterion_feedback,
                overall_feedback,
                rubric_dict,
                scored_at,
                scorer_workflow
            )
            assessment_dicts.append(full_assessment_dict(assessment))

        return assessment_dicts

    except InvalidRubric as ex:
        error_message = "The rubric definition is not valid."
        logger.exception(error_message)
        raise StaffAssessmentRequestError(error_message) from ex
    except InvalidRubricSelection as ex:
        error_message = "Invalid options were selected in the rubric."
        logger.warning(error_message, exc_info=True)
        raise StaffAssessmentRequestError(error_message) from ex
    except DatabaseError as ex:
        error_message = (
            "An error occurred while creating an assessment by the scorer with this ID: {}"
        ).format(scorer_id)
        logger.exception(error_message)
        raise StaffAssessmentInternalError(error_message) from ex



"""
Errors for the peer assessment.
"""
from .base import AssessmentError


class PeerAssessmentError(AssessmentError):
    """Generic Peer Assessment Error

    Raised when an error occurs while processing a request related to the
    Peer Assessment Workflow.

    """


class PeerAssessmentRequestError(PeerAssessmentError):
    """Error indicating insufficient or incorrect parameters in the request.

    Raised when the request does not contain enough information, or incorrect
    information which does not allow the request to be processed.

    """


class PeerAssessmentWorkflowError(PeerAssessmentError):
    """Error indicating a step in the workflow cannot be completed,

    Raised when the action taken cannot be completed in the workflow. This can
    occur based on parameters specific to the Submission, User, or Peer Scorers.

    """


class PeerAssessmentInternalError(PeerAssessmentError):
    """Error indicating an internal problem independent of API use.

    Raised when an internal error has occurred. This should be independent of
    the actions or parameters given to the API.

    """



"""JSON handlers for the old-style ORA UI"""

import logging

from xblock.core import XBlock

from openassessment.xblock.staff_area_mixin import require_course_staff
from openassessment.xblock.ui_mixins.legacy.peer_assessments.actions import peer_assess
from openassessment.xblock.ui_mixins.legacy.self_assessments.actions import self_assess
from openassessment.xblock.ui_mixins.legacy.staff_assessments.actions import (
    do_staff_assessment,
    staff_assess,
)
from openassessment.xblock.ui_mixins.legacy.student_training.actions import training_assess

from openassessment.xblock.ui_mixins.legacy.submissions.actions import (
    submit,
    save_submission,
)
from openassessment.xblock.utils.data_conversion import verify_assessment_parameters
from openassessment.xblock.ui_mixins.legacy.submissions.file_actions import (
    save_files_descriptions,
    upload_url,
    download_url,
    remove_uploaded_file,
)

logger = logging.getLogger(__name__)


class LegacyHandlersMixin:
    """
    Exposes actions (@XBlock.json_handlers) used in our legacy ORA UI
    """

    # Submissions

    @XBlock.json_handler
    def submit(self, data, suffix=""):  # pylint: disable=unused-argument
        """Submit a response for the student provided in data['submission']"""
        return submit(self.config_data, self.submission_data, data)

    @XBlock.json_handler
    def save_submission(self, data, suffix=""):  # pylint: disable=unused-argument
        """Save a draft response for the student under data['submission']"""
        return save_submission(self.config_data, self.submission_data, data)

    # File uploads

    @XBlock.json_handler
    def save_files_descriptions(self, data, suffix=""):  # pylint: disable=unused-argument
        return save_files_descriptions(self.config_data, self.submission_data, data)

    @XBlock.json_handler
    def upload_url(self, data, suffix=""):  # pylint: disable=unused-argument
        return upload_url(self.config_data, self.submission_data, data)

    @XBlock.json_handler
    def download_url(self, data, suffix=""):  # pylint: disable=unused-argument
        return download_url(self.submission_data, data)

    @XBlock.json_handler
    def remove_uploaded_file(self, data, suffix=""):  # pylint: disable=unused-argument
        return remove_uploaded_file(self.config_data, self.submission_data, data)

    # Assessments

    @XBlock.json_handler
    @verify_assessment_parameters
    def peer_assess(self, data, suffix=""):  # pylint: disable=unused-argument
        return peer_assess(self.api_data, data)

    @XBlock.json_handler
    @verify_assessment_parameters
    def self_assess(self, data, suffix=""):  # pylint: disable=unused-argument
        return self_assess(self.api_data, data)

    @XBlock.json_handler
    def training_assess(self, data, suffix=""):  # pylint: disable=unused-argument
        return training_assess(self.api_data, data)

    @XBlock.json_handler
    @require_course_staff("STUDENT_INFO")
    @verify_assessment_parameters
    def staff_assess(self, data, suffix=""):  # pylint: disable=unused-argument
        return staff_assess(self.api_data, data)

    # NOTE - Temporary surfacing
    def do_staff_assessment(self, data):
        return do_staff_assessment(self.api_data, data)

    # Utils

    @XBlock.json_handler
    def get_student_username(self, data, suffix=""):  # pylint: disable=unused-argument
        """
        Gets the username of the current student for use in team lookup.
        """
        anonymous_id = self.xmodule_runtime.anonymous_student_id
        return {"username": self.get_username(anonymous_id)}



"""
Serializers are created to ensure models do not have to be accessed outside the
scope of the ORA2 APIs.
"""


from rest_framework import serializers

from openassessment.workflow.models import AssessmentWorkflow, AssessmentWorkflowCancellation, TeamAssessmentWorkflow


class AssessmentWorkflowSerializer(serializers.ModelSerializer):
    """
    Serialize a AssessmentWorkflow' model.
    """
    score = serializers.ReadOnlyField(required=False)
    assessment_score_priority = serializers.SerializerMethodField()

    def get_assessment_score_priority(self, obj):
        return obj.ASSESSMENT_SCORE_PRIORITY

    class Meta:
        model = AssessmentWorkflow
        fields = (
            'submission_uuid',
            'status',
            'created',
            'modified',

            # Computed
            'score',
            'assessment_score_priority',
        )


class TeamAssessmentWorkflowSerializer(serializers.ModelSerializer):
    """
    Serialize a TeamAssessmentWorkflow model.
    """
    score = serializers.ReadOnlyField(required=False)

    class Meta:
        model = TeamAssessmentWorkflow
        fields = (
            'team_submission_uuid',
            'submission_uuid',
            'status',
            'created',
            'modified',

            # Computed
            'score'
        )


class AssessmentWorkflowCancellationSerializer(serializers.ModelSerializer):
    """
    Serialize a `AssessmentWorkflowCancellation` model.
    """

    class Meta:
        model = AssessmentWorkflowCancellation
        fields = (
            'comments',
            'cancelled_by_id',
            'created_at',
        )



"""
Actions related to submissions
"""

import json
import logging

from openassessment.workflow.errors import AssessmentWorkflowError
from openassessment.xblock.utils.data_conversion import (
    prepare_submission_for_serialization,
)
from openassessment.xblock.apis.submissions.errors import (
    EmptySubmissionError,
    NoTeamToCreateSubmissionForError,
)
from openassessment.xblock.utils.validation import validate_submission

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def submit(block_config, submission_info, data):
    """
    Place the submission text into Openassessment system

    Allows submission of new responses.  Performs basic workflow validation
    on any new submission to ensure it is acceptable to receive a new
    response at this time.

    Args:
        data (dict): Data may contain two attributes: submission and
            file_urls. submission is the response from the student which
            should be stored in the Open Assessment system. file_urls is the
            path to a related file for the submission. file_urls is optional.
        suffix (str): Not used in this handler.

    Returns:
        (tuple | [tuple]): Returns the status (boolean) of this request, the
            associated status tag (str), and status text (unicode).
            This becomes an array of similarly structured tuples in the event
            of a team submission, one entry per student entry.
    """
    # Import is placed here to avoid model import at project startup.
    from submissions import api

    if "submission" not in data:
        return (
            False,
            "EBADARGS",
            block_config.translate('"submission" required to submit answer.'),
        )

    status = False
    student_sub_data = data["submission"]
    success, msg = validate_submission(
        student_sub_data,
        block_config.prompts,
        block_config.translate,
        block_config.text_response,
    )
    if not success:
        return (False, "EBADARGS", msg)

    student_item_dict = block_config.student_item_dict

    # Short-circuit if no user is defined (as in Studio Preview mode)
    # Since students can't submit, they will never be able to progress in the workflow
    if block_config.in_studio_preview:
        return (
            False,
            "ENOPREVIEW",
            block_config.translate("To submit a response, view this component in Preview or Live mode."),
        )

    status_tag = "ENOMULTI"  # It is an error to submit multiple times for the same item
    status_text = block_config.translate("Multiple submissions are not allowed.")

    if not submission_info.has_submitted:
        try:
            # a submission for a team generates matching submissions for all members
            if block_config.is_team_assignment():
                submission = submission_info.create_team_submission(student_item_dict, student_sub_data)
            else:
                submission = submission_info.create_submission(student_item_dict, student_sub_data)
            return _create_submission_response(submission)

        except api.SubmissionRequestError as err:
            # Handle the case of an answer that's too long as a special case,
            # so we can display a more specific error message.
            # Although we limit the number of characters the user can
            # enter on the client side, the submissions API uses the JSON-serialized
            # submission to calculate length.  If each character submitted
            # by the user takes more than 1 byte to encode (for example, double-escaped
            # newline characters or non-ASCII unicode), then the user might
            # exceed the limits set by the submissions API.  In that case,
            # we display an error message indicating that the answer is too long.
            answer_too_long = any(
                "maximum answer size exceeded" in answer_err.lower()
                for answer_err in err.field_errors.get("answer", [])
            )
            if answer_too_long:
                logger.exception(f"Response exceeds maximum allowed size: {student_item_dict}")
                status_tag = "EANSWERLENGTH"
                max_size = f"({int(api.Submission.MAXSIZE / 1024)} KB)"
                base_error = block_config.translate("Response exceeds maximum allowed size.")
                extra_info = block_config.translate(
                    "Note: if you have a spellcheck or grammar check browser extension, "
                    "try disabling, reloading, and reentering your response before submitting."
                )
                status_text = f"{base_error} {max_size} {extra_info}"
            else:
                msg = (
                    "The submissions API reported an invalid request error "
                    "when submitting a response for the user: {student_item}"
                ).format(student_item=student_item_dict)
                logger.exception(msg)
                status_tag = "EBADFORM"
                status_text = msg
        except EmptySubmissionError:
            msg = (
                "Attempted to submit submission for user {student_item}, " "but submission contained no content."
            ).format(student_item=student_item_dict)
            logger.exception(msg)
            status_tag = "EEMPTYSUB"
            status_text = block_config.translate(
                "Submission cannot be empty. Please refresh the page and try again."
            )
        except (
            api.SubmissionError,
            AssessmentWorkflowError,
            NoTeamToCreateSubmissionForError,
        ):
            msg = ("An unknown error occurred while submitting " "a response for the user: {student_item}").format(
                student_item=student_item_dict
            )
            logger.exception(msg)
            status_tag = "EUNKNOWN"
            status_text = block_config.translate("API returned unclassified exception.")

    # error cases fall through to here
    return status, status_tag, status_text


def save_submission(block_config, submission_info, data):
    """
    Save the current student's response submission.
    If the student already has a response saved, this will overwrite it.

    Args:
        data (dict): Data should have a single key 'submission' that contains
            the text of the student's response. Optionally, the data could
            have a 'file_urls' key that is the path to an associated file for
            this submission.
        suffix (str): Not used.

    Returns:
        dict: Contains a bool 'success' and unicode string 'msg'.
    """
    if "submission" in data:
        student_sub_data = data["submission"]
        success, msg = validate_submission(
            student_sub_data,
            block_config.prompts,
            block_config.translate,
            block_config.text_response,
        )
        if not success:
            return {"success": False, "msg": msg}
        try:
            submission_info.saved_response = json.dumps(prepare_submission_for_serialization(student_sub_data))
            submission_info.has_saved = True

            # Emit analytics event...
            block_config.publish_event(
                "openassessmentblock.save_submission",
                {"saved_response": submission_info.saved_response},
            )
        except Exception:  # pylint: disable=broad-except
            return {
                "success": False,
                "msg": block_config.translate("Please contact support staff."),
            }
        else:
            return {"success": True, "msg": ""}
    else:
        return {
            "success": False,
            "msg": block_config.translate("Submission data missing. Please contact support staff."),
        }


def _create_submission_response(submission):
    """
    Wrap submission info for return to client

    Returns:
        (tuple): True (indicates success), student item, attempt number
    """
    status = True
    status_tag = submission.get("student_item")
    status_text = submission.get("attempt_number")

    return (status, status_tag, status_text)



"""
Serializers common to all assessment types.
"""


from copy import deepcopy
import logging

from rest_framework import serializers
from rest_framework.fields import DateTimeField, IntegerField

from django.core.cache import cache

from openassessment.assessment.models import Assessment, AssessmentPart, Criterion, CriterionOption, Rubric

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class InvalidRubric(Exception):
    """This can be raised during the deserialization process."""
    def __init__(self, errors):
        Exception.__init__(self, repr(errors))
        self.errors = deepcopy(errors)


class CriterionOptionSerializer(serializers.ModelSerializer):
    """Serializer for :class:`CriterionOption`"""

    # Django Rest Framework v3 no longer requires `PositiveIntegerField`s
    # to be positive by default, so we need to explicitly set the `min_value`
    # on the serializer field.
    points = IntegerField(min_value=0)

    class Meta:
        model = CriterionOption
        fields = ('order_num', 'points', 'name', 'label', 'explanation')


class CriterionSerializer(serializers.ModelSerializer):
    """Serializer for :class:`Criterion`"""
    options = CriterionOptionSerializer(required=True, many=True)

    class Meta:
        model = Criterion
        fields = ('order_num', 'name', 'label', 'prompt', 'options', 'points_possible')


class RubricSerializer(serializers.ModelSerializer):
    """Serializer for :class:`Rubric`."""
    criteria = CriterionSerializer(required=True, many=True)

    class Meta:
        model = Rubric
        fields = ('id', 'content_hash', 'structure_hash', 'criteria', 'points_possible')

    def validate_criteria(self, value):
        """Make sure we have at least one Criterion in the Rubric."""
        if not value:
            raise serializers.ValidationError("Must have at least one criterion")
        return value

    @classmethod
    def serialized_from_cache(cls, rubric, local_cache=None):
        """For a given `Rubric` model object, return a serialized version.

        This method will attempt to use the cache if possible, first looking at
        the `local_cache` dict you can pass in, and then looking at whatever
        Django cache is configured.

        Args:
            rubric (Rubric): The Rubric model to get the serialized form of.
            local_cach (dict): Mapping of `rubric.content_hash` to serialized
                rubric dictionary. We include this so that we can call this
                method in a loop.

        Returns:
            dict: `Rubric` fields as a dictionary, with `criteria` and `options`
                relations followed.
        """
        # Optional local cache you can send in (for when you're calling this
        # in a loop).
        local_cache = local_cache or {}

        # Check our in-memory cache...
        if rubric.content_hash in local_cache:
            return local_cache[rubric.content_hash]

        # Check the external cache (e.g. memcached)
        rubric_dict_cache_key = (
            "RubricSerializer.serialized_from_cache.{}"
            .format(rubric.content_hash)
        )
        rubric_dict = cache.get(rubric_dict_cache_key)
        if rubric_dict:
            local_cache[rubric.content_hash] = rubric_dict
            return rubric_dict

        # Grab it from the database
        rubric_dict = RubricSerializer(rubric).data
        cache.set(rubric_dict_cache_key, rubric_dict)
        local_cache[rubric.content_hash] = rubric_dict

        return rubric_dict

    def create(self, validated_data):
        """
        Create the rubric model, including its nested models.

        Args:
            validated_data (dict): Dictionary of validated data for the rubric,
                including nested Criterion and CriterionOption data.

        Returns:
            Rubric
        """
        criteria_data = validated_data.pop("criteria")
        rubric = Rubric.objects.create(**validated_data)

        # Create each nested criterion in the rubric, linking it to the rubric
        for criterion_dict in criteria_data:
            options_data = criterion_dict.pop("options")
            criterion = Criterion.objects.create(rubric=rubric, **criterion_dict)

            # Create each option in the criterion, linking it to the criterion
            CriterionOption.objects.bulk_create(
                CriterionOption(criterion=criterion, **option_dict)
                for option_dict in options_data
            )

        return rubric


class AssessmentPartSerializer(serializers.ModelSerializer):
    """Serializer for :class:`AssessmentPart`."""

    class Meta:
        model = AssessmentPart
        fields = ('option', 'criterion', 'feedback')


class AssessmentSerializer(serializers.ModelSerializer):
    """Simplified serializer for :class:`Assessment` that's lighter on the DB."""

    # Django Rest Framework v3 uses the Django setting `DATETIME_FORMAT`
    # when serializing datetimes.  This differs from v2, which always
    # returned a datetime.  To preserve the old behavior, we explicitly
    # set `format` to None.
    # http://www.django-rest-framework.org/api-guide/fields/#datetimefield
    scored_at = DateTimeField(format=None, required=False)

    class Meta:
        model = Assessment
        fields = (
            'submission_uuid',
            'rubric',
            'scored_at',
            'scorer_id',
            'score_type',
            'feedback',
        )


def serialize_assessments(assessments_qset):
    assessments = list(assessments_qset.select_related("rubric"))
    rubric_cache = {}

    return [
        full_assessment_dict(
            assessment,
            RubricSerializer.serialized_from_cache(
                assessment.rubric, rubric_cache
            )
        )
        for assessment in assessments
    ]


def full_assessment_dict(assessment, rubric_dict=None):
    """
    Return a dict representation of the Assessment model, including nested
    assessment parts. We do some of the serialization ourselves here instead
    of relying on the Django REST Framework serializers. This is for performance
    reasons -- we have a cached rubric easily available, and we don't want to
    follow all the DB relations from assessment -> assessment part -> option ->
    criterion.

    Args:
        assessment (Assessment): The Assessment model to serialize

    Returns:
        dict with keys 'rubric' (serialized Rubric model) and 'parts' (serialized assessment parts)
    """
    assessment_cache_key = "assessment.full_assessment_dict.{}.{}.{}".format(
        assessment.id, assessment.submission_uuid, assessment.scored_at.isoformat()
    )
    assessment_dict = cache.get(assessment_cache_key)
    if assessment_dict:
        return assessment_dict

    assessment_dict = AssessmentSerializer(assessment).data
    if not rubric_dict:
        rubric_dict = RubricSerializer.serialized_from_cache(assessment.rubric)

    assessment_dict["rubric"] = rubric_dict

    # This part looks a little goofy, but it's in the name of saving dozens of
    # SQL lookups. The rubric_dict has the entire serialized output of the
    # `Rubric`, its child `Criterion` and grandchild `CriterionOption`. This
    # includes calculated things like `points_possible` which aren't actually in
    # the DB model. Instead of invoking the serializers for `Criterion` and
    # `CriterionOption` again, we simply index into the places we expect them to
    # be from the big, saved `Rubric` serialization.
    parts = []
    for part in assessment.parts.order_by('criterion__order_num').all().select_related("criterion", "option"):
        criterion_dict = dict(rubric_dict["criteria"][part.criterion.order_num])
        options_dict = None
        if part.option is not None:
            options_dict = criterion_dict["options"][part.option.order_num]
            options_dict["criterion"] = criterion_dict
        parts.append({
            "option": options_dict,
            "criterion": criterion_dict,
            "feedback": part.feedback
        })

    # Now manually built up the dynamically calculated values on the
    # `Assessment` so we can again avoid DB calls.
    assessment_dict["parts"] = parts
    assessment_dict["points_earned"] = sum(
        part_dict["option"]["points"]
        if part_dict["option"] is not None else 0
        for part_dict in parts
    )
    assessment_dict["points_possible"] = rubric_dict["points_possible"]
    assessment_dict["id"] = assessment.id

    cache.set(assessment_cache_key, assessment_dict)

    return assessment_dict


def rubric_from_dict(rubric_dict):
    """Given a dict of rubric information, return the corresponding Rubric

    This will create the Rubric and its children if it does not exist already.

    Sample data (one criterion, two options)::

        {
          "prompts": [{"description": "Create a plan to deliver ora2!"}],
          "criteria": [
            {
              "order_num": 0,
              "name": "realistic",
              "prompt": "Is the deadline realistic?",
              "options": [
                {
                  "order_num": 0,
                  "points": 0,
                  "name": "No",
                  "explanation": "We need more time!"
                },
                {
                  "order_num": 1,
                  "points": 2,
                  "name": "Yes",
                  "explanation": "We got this."
                },
              ]
            }
          ]
        }

    """
    rubric_dict = deepcopy(rubric_dict)

    # Calculate the hash based on the rubric content...
    content_hash = Rubric.content_hash_from_dict(rubric_dict)

    try:
        rubric = Rubric.objects.get(content_hash=content_hash)
    except Rubric.DoesNotExist as ex:
        rubric_dict["content_hash"] = content_hash
        rubric_dict["structure_hash"] = Rubric.structure_hash_from_dict(rubric_dict)
        for crit_idx, criterion in enumerate(rubric_dict.get("criteria", {})):
            if "order_num" not in criterion:
                criterion["order_num"] = crit_idx
            for opt_idx, option in enumerate(criterion.get("options", {})):
                if "order_num" not in option:
                    option["order_num"] = opt_idx

        rubric_serializer = RubricSerializer(data=rubric_dict)
        if not rubric_serializer.is_valid():
            raise InvalidRubric(rubric_serializer.errors) from ex
        rubric = rubric_serializer.save()

    return rubric



"""
Public interface for student training:

* Staff can create assessments for example responses.
* Students assess an example response, then compare the scores
    they gave to to the instructor's assessment.

"""


import logging

from django.db import DatabaseError
from django.utils.translation import gettext as _

from submissions import api as sub_api

from openassessment.assessment.errors import StudentTrainingInternalError, StudentTrainingRequestError
from openassessment.assessment.models import InvalidRubricSelection, StudentTrainingWorkflow
from openassessment.assessment.serializers import (InvalidRubric, InvalidTrainingExample, deserialize_training_examples,
                                                   serialize_training_example, validate_training_example_format)

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def submitter_is_finished(submission_uuid, training_requirements):
    """
    Check whether the student has correctly assessed
    all the training example responses.

    Args:
        submission_uuid (str): The UUID of the student's submission.
        training_requirements (dict): Must contain "num_required" indicating
            the number of examples the student must assess.

    Returns:
        bool

    Raises:
        StudentTrainingRequestError

    """
    if training_requirements is None:
        return False

    try:
        num_required = int(training_requirements['num_required'])
    except KeyError as ex:
        raise StudentTrainingRequestError('Requirements dict must contain "num_required" key') from ex
    except ValueError as ex:
        raise StudentTrainingRequestError('Number of requirements must be an integer') from ex

    try:
        workflow = StudentTrainingWorkflow.objects.get(submission_uuid=submission_uuid)
    except StudentTrainingWorkflow.DoesNotExist:
        return False
    else:
        return workflow.num_completed >= num_required


def on_start(submission_uuid):
    """
    Creates a new student training workflow.

    This function should be called to indicate that a submission has entered the
    student training workflow part of the assessment process.

    Args:
        submission_uuid (str): The submission UUID for the student that is
            initiating training.

    Returns:
        None

    Raises:
        StudentTrainingInternalError: Raised when an error occurs persisting the
            Student Training Workflow
    """
    try:
        StudentTrainingWorkflow.create_workflow(submission_uuid)
    except Exception as ex:
        msg = (
            "An internal error has occurred while creating the learner "
            "training workflow for submission UUID {}".format(submission_uuid)
        )
        logger.exception(msg)
        raise StudentTrainingInternalError(msg) from ex


def validate_training_examples(rubric, examples):
    """
    Validate that the training examples match the rubric.

    Args:
        rubric (dict): Serialized rubric model.
        examples (list): List of serialized training examples.

    Returns:
        list of errors (unicode)

    Raises:
        StudentTrainingRequestError
        StudentTrainingInternalError

    Example usage:

        >>> options = [
        >>>     {
        >>>         "order_num": 0,
        >>>         "name": "poor",
        >>>         "explanation": "Poor job!",
        >>>         "points": 0,
        >>>     },
        >>>     {
        >>>         "order_num": 1,
        >>>         "name": "good",
        >>>         "explanation": "Good job!",
        >>>         "points": 1,
        >>>     },
        >>>     {
        >>>         "order_num": 2,
        >>>         "name": "excellent",
        >>>         "explanation": "Excellent job!",
        >>>         "points": 2,
        >>>     },
        >>> ]
        >>>
        >>> rubric = {
        >>>     "prompts": [{"description": "Write an essay!"}],
        >>>     "criteria": [
        >>>         {
        >>>             "order_num": 0,
        >>>             "name": "vocabulary",
        >>>             "prompt": "How varied is the vocabulary?",
        >>>             "options": options
        >>>         },
        >>>         {
        >>>             "order_num": 1,
        >>>             "name": "grammar",
        >>>             "prompt": "How correct is the grammar?",
        >>>             "options": options
        >>>         }
        >>>     ]
        >>> }
        >>>
        >>> examples = [
        >>>     {
        >>>         'answer': {'parts': [{'text': u'Lorem ipsum'}]},
        >>>         'options_selected': {
        >>>             'vocabulary': 'good',
        >>>             'grammar': 'excellent'
        >>>         }
        >>>     },
        >>>     {
        >>>         'answer': {'parts': [{'text': u'Doler'}]},
        >>>         'options_selected': {
        >>>             'vocabulary': 'good',
        >>>             'grammar': 'poor'
        >>>         }
        >>>     }
        >>> ]
        >>>
        >>> errors = validate_training_examples(rubric, examples)

    """
    errors = []

    # Construct a list of valid options for each criterion
    try:
        criteria_options = {
            str(criterion['name']): [
                str(option['name'])
                for option in criterion['options']
            ]
            for criterion in rubric['criteria']
        }
    except (ValueError, KeyError):
        logger.warning("Could not parse serialized rubric", exc_info=True)
        return [_("Could not parse serialized rubric")]

    # Check that at least one criterion in the rubric has options
    # If this is not the case (that is, if all rubric criteria are written feedback only),
    # then it doesn't make sense to do student training.
    criteria_without_options = [
        criterion_name
        for criterion_name, criterion_option_list in criteria_options.items()
        if len(criterion_option_list) == 0
    ]
    if not (set(criteria_options) - set(criteria_without_options)):
        return [_(
            "If your assignment includes a learner training step, "
            "the rubric must have at least one criterion, "
            "and that criterion must have at least one option."
        )]

    # Check each example
    for order_num, example_dict in enumerate(examples, start=1):

        # Check the structure of the example dict
        is_format_valid, format_errors = validate_training_example_format(example_dict)
        if not is_format_valid:
            format_errors = [
                _("Example {example_number} has a validation error: {error}").format(
                    example_number=order_num, error=error
                )
                for error in format_errors
            ]
            errors.extend(format_errors)
        else:
            # Check each selected option in the example (one per criterion)
            options_selected = example_dict['options_selected']
            for criterion_name, option_name in options_selected.items():
                if criterion_name in criteria_options:
                    valid_options = criteria_options[criterion_name]
                    if option_name not in valid_options:
                        msg = _(
                            "Example {example_number} has an invalid option "
                            "for \"{criterion_name}\": \"{option_name}\""
                        ).format(
                            example_number=order_num,
                            criterion_name=criterion_name,
                            option_name=option_name
                        )
                        errors.append(msg)
                else:
                    msg = _(
                        "Example {example_number} has an extra option "
                        "for \"{criterion_name}\""
                    ).format(
                        example_number=order_num,
                        criterion_name=criterion_name
                    )
                    errors.append(msg)

            # Check for missing criteria
            # Ignore options
            all_example_criteria = set(list(options_selected.keys()) + criteria_without_options)
            for missing_criterion in set(criteria_options.keys()) - all_example_criteria:
                msg = _(
                    "Example {example_number} is missing an option "
                    "for \"{criterion_name}\""
                ).format(
                    example_number=order_num,
                    criterion_name=missing_criterion
                )
                errors.append(msg)

    return errors


def get_num_completed(submission_uuid):
    """
    Get the number of training examples the student has assessed successfully.

    Args:
        submission_uuid (str): The UUID of the student's submission.

    Returns:
        int: The number of completed training examples

    Raises:
        StudentTrainingInternalError

    Example usage:
        >>> get_num_completed("5443ebbbe2297b30f503736e26be84f6c7303c57")
        2

    """
    try:
        try:
            workflow = StudentTrainingWorkflow.objects.get(submission_uuid=submission_uuid)
        except StudentTrainingWorkflow.DoesNotExist:
            return 0
        else:
            return workflow.num_completed
    except DatabaseError as ex:
        msg = (
            "An unexpected error occurred while "
            "retrieving the learner training workflow status for submission UUID {}"
        ).format(submission_uuid)
        logger.exception(msg)
        raise StudentTrainingInternalError(msg) from ex


def get_training_example(submission_uuid, rubric, examples):
    """
    Retrieve a training example for the student to assess.
    This will implicitly create a workflow for the student if one does not yet exist.

    NOTE: We include the rubric in the returned dictionary to handle
    the case in which the instructor changes the rubric definition
    while the student is assessing the training example.  Once a student
    starts on a training example, the student should see the same training
    example consistently.  However, the next training example the student
    retrieves will use the updated rubric.

    Args:
        submission_uuid (str): The UUID of the student's submission.
        rubric (dict): Serialized rubric model.
        examples (list): List of serialized training examples.

    Returns:
        dict: The training example with keys "answer", "rubric", and "options_selected".
        If no training examples are available (the student has already assessed every example,
            or no examples are defined), returns None.

    Raises:
        StudentTrainingInternalError

    Example usage:

        >>> examples = [
        >>>     {
        >>>         'answer': {
        >>>             'parts': {
        >>>                 [
        >>>                     {'text:' 'Answer part 1'},
        >>>                     {'text:' 'Answer part 2'},
        >>>                     {'text:' 'Answer part 3'}
        >>>                 ]
        >>>             }
        >>>         },
        >>>         'options_selected': {
        >>>             'vocabulary': 'good',
        >>>             'grammar': 'poor'
        >>>         }
        >>>     }
        >>> ]
        >>>
        >>> get_training_example("5443ebbbe2297b30f503736e26be84f6c7303c57", rubric, examples)
        {
             'answer': {
                 'parts': {
                     [
                         {'text:' 'Answer part 1'},
                         {'text:' 'Answer part 2'},
                         {'text:' 'Answer part 3'}
                     ]
                 }
             },
            'rubric': {
                "prompts": [
                    {"description": "Prompt 1"},
                    {"description": "Prompt 2"},
                    {"description": "Prompt 3"}
                ],
                "criteria": [
                    {
                        "order_num": 0,
                        "name": "vocabulary",
                        "prompt": "How varied is the vocabulary?",
                        "options": options
                    },
                    {
                        "order_num": 1,
                        "name": "grammar",
                        "prompt": "How correct is the grammar?",
                        "options": options
                    }
                ],
            },
            'options_selected': {
                'vocabulary': 'good',
                'grammar': 'excellent'
            }
        }

    """
    try:
        # Validate the training examples
        errors = validate_training_examples(rubric, examples)
        if errors:
            msg = (
                "Training examples do not match the rubric (submission UUID is {uuid}): {errors}"
            ).format(uuid=submission_uuid, errors="\n".join(errors))
            raise StudentTrainingRequestError(msg)

        # Get or create the workflow
        workflow = StudentTrainingWorkflow.get_workflow(submission_uuid=submission_uuid)
        if not workflow:
            raise StudentTrainingRequestError(
                f"No learner training workflow found for submission {submission_uuid}"
            )

        # Get or create the training examples
        examples = deserialize_training_examples(examples, rubric)

        # Pick a training example that the student has not yet completed
        # If the student already started a training example, then return that instead.
        next_example = workflow.next_training_example(examples)
        return None if next_example is None else serialize_training_example(next_example)
    except (InvalidRubric, InvalidRubricSelection, InvalidTrainingExample) as ex:
        logger.exception("Could not deserialize training examples for submission UUID %s", submission_uuid)
        raise StudentTrainingRequestError(ex) from ex
    except sub_api.SubmissionNotFoundError as ex:
        logger.exception("Could not retrieve the submission with UUID %s", submission_uuid)
        raise StudentTrainingRequestError(msg) from ex
    except DatabaseError as ex:
        msg = "Could not retrieve a training example for the learner with submission UUID %s"
        logger.exception(msg, submission_uuid)
        raise StudentTrainingInternalError(msg % submission_uuid) from ex


def assess_training_example(submission_uuid, options_selected, update_workflow=True):
    """
    Assess a training example and update the workflow.

    This must be called *after* `get_training_example()`.

    Args:
        submission_uuid (str): The UUID of the student's submission.
        options_selected (dict): The options the student selected.

    Keyword Arguments:
        update_workflow (bool): If true, mark the current item complete
            if the student has assessed the example correctly.

    Returns:
        corrections (dict): Dictionary containing the correct
            options for criteria the student scored incorrectly.

    Raises:
        StudentTrainingRequestError
        StudentTrainingInternalError

    Example usage:

        >>> options_selected = {
        >>>     'vocabulary': 'good',
        >>>     'grammar': 'excellent'
        >>> }
        >>> assess_training_example("5443ebbbe2297b30f503736e26be84f6c7303c57", options_selected)
        {'grammar': 'poor'}

    """
    # Find a workflow for the student
    try:
        workflow = StudentTrainingWorkflow.objects.get(submission_uuid=submission_uuid)

        # Find the item the student is currently working on
        item = workflow.current_item
        if item is None:
            msg = (
                "No items are available in the learner training workflow associated with "
                "submission UUID {}"
            ).format(submission_uuid)
            raise StudentTrainingRequestError(msg)

        # Check the student's scores against the staff's scores.
        corrections = item.check_options(options_selected)

        # Mark the item as complete if the student's selection
        # matches the instructor's selection
        if update_workflow and not corrections:
            item.mark_complete()
        return corrections
    except StudentTrainingWorkflow.DoesNotExist as ex:
        msg = f"Could not find learner training workflow for submission UUID {submission_uuid}"
        raise StudentTrainingRequestError(msg) from ex
    except DatabaseError as ex:
        msg = (
            "An error occurred while comparing the learner's assessment "
            "to the training example.  The submission UUID for the learner is {}"
        ).format(submission_uuid)
        logger.exception(msg)
        raise StudentTrainingInternalError(msg) from ex



# pylint: skip-file


from django.db import migrations, models
import django.utils.timezone

import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AssessmentWorkflow',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('status', model_utils.fields.StatusField(default='peer', max_length=100, verbose_name='status', no_check_for_status=True, choices=[('peer', 'peer'), ('ai', 'ai'), ('self', 'self'), ('training', 'training'), ('waiting', 'waiting'), ('done', 'done'), ('cancelled', 'cancelled')])),
                ('status_changed', model_utils.fields.MonitorField(default=django.utils.timezone.now, verbose_name='status changed', monitor='status')),
                ('submission_uuid', models.CharField(unique=True, max_length=36, db_index=True)),
                ('uuid', models.UUIDField(db_index=True, unique=True, editable=False, blank=True)),
                ('course_id', models.CharField(max_length=255, db_index=True)),
                ('item_id', models.CharField(max_length=255, db_index=True)),
            ],
            options={
                'ordering': ['-created'],
            },
        ),
        migrations.CreateModel(
            name='AssessmentWorkflowCancellation',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('comments', models.TextField(max_length=10000)),
                ('cancelled_by_id', models.CharField(max_length=40, db_index=True)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, db_index=True)),
                ('workflow', models.ForeignKey(related_name='cancellations', to='workflow.AssessmentWorkflow', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['created_at', 'id'],
            },
        ),
        migrations.CreateModel(
            name='AssessmentWorkflowStep',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=20)),
                ('submitter_completed_at', models.DateTimeField(default=None, null=True)),
                ('assessment_completed_at', models.DateTimeField(default=None, null=True)),
                ('order_num', models.PositiveIntegerField()),
                ('workflow', models.ForeignKey(related_name='steps', to='workflow.AssessmentWorkflow', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['workflow', 'order_num'],
            },
        ),
    ]



"""API Data wrapper for exposed APIs within ORA XBlock"""
from openassessment.xblock.apis.ora_config_api import ORAConfigAPI
from openassessment.xblock.apis.submissions.submissions_api import SubmissionAPI
from openassessment.xblock.apis.workflow_api import WorkflowAPI
from openassessment.xblock.apis.assessments.peer_assessment_api import PeerAssessmentAPI
from openassessment.xblock.apis.assessments.self_assessment_api import SelfAssessmentAPI
from openassessment.xblock.apis.assessments.staff_assessment_api import (
    StaffAssessmentAPI,
)
from openassessment.xblock.apis.assessments.student_training_api import (
    StudentTrainingAPI,
)


class ORADataAccessor:
    def __init__(self, block):
        self._block = block

    @property
    def config_data(self):
        return ORAConfigAPI(self._block)

    @property
    def submission_data(self):
        return SubmissionAPI(self._block)

    @property
    def workflow_data(self):
        return WorkflowAPI(self._block)

    @property
    def self_assessment_data(self):
        return SelfAssessmentAPI(self._block)

    @property
    def staff_assessment_data(self):
        return StaffAssessmentAPI(self._block)

    @property
    def student_training_data(self):
        return StudentTrainingAPI(self._block)

    def peer_assessment_data(self, continue_grading=False):
        return PeerAssessmentAPI(self._block, continue_grading)



"""
API endpoints for enhanced staff grader
"""
from functools import wraps
import logging

from django.db.models import Case, OuterRef, Prefetch, Subquery, Value, When
from django.db.models.fields import CharField
from xblock.core import XBlock
from xblock.exceptions import JsonHandlerError

from submissions.api import get_student_ids_by_submission_uuid, get_submission
from submissions.errors import SubmissionInternalError, SubmissionNotFoundError, SubmissionRequestError, SubmissionError
from submissions.team_api import get_team_ids_by_team_submission_uuid, get_team_submission

from openassessment.assessment.models.base import Assessment, AssessmentPart
from openassessment.assessment.models.staff import StaffWorkflow, TeamStaffWorkflow
from openassessment.data import map_anonymized_ids_to_usernames, OraSubmissionAnswerFactory, VersionNotFoundException
from openassessment.staffgrader.errors.submission_lock import SubmissionLockContestedError
from openassessment.staffgrader.models.submission_lock import SubmissionGradingLock
from openassessment.staffgrader.serializers import (
    AssessmentSerializer,
    MissingContextException,
    SubmissionDetailFileSerilaizer,
    SubmissionListSerializer,
    SubmissionLockSerializer,
    TeamSubmissionListSerializer,
)
from openassessment.xblock.staff_area_mixin import require_course_staff


log = logging.getLogger(__name__)


def require_submission_uuid(validate=True):
    """
    Unpacks and passes submission_uuid from request to handler function.

    params:
    - validate: Whether or not to check submissions to see if this is a real submission UUID or not. Default True

    Raises:
    - 400 if the submission_uuid was not provided or was incorrectly formatted
    - 404 if the submission_uuid wasn't found in submissions
    - 500 for errors with submissions or general exceptions
    """
    def decorator(handler):
        @wraps(handler)
        def wrapped_handler(self, data, suffix=""):  # pylint: disable=unused-argument
            submission_uuid = data.get('submission_uuid', None)
            if not submission_uuid:
                raise JsonHandlerError(400, "Body must contain a submission_uuid")
            if validate:
                try:
                    if self.is_team_assignment():
                        get_team_submission(submission_uuid)
                    else:
                        get_submission(submission_uuid)
                except SubmissionNotFoundError as exc:
                    raise JsonHandlerError(404, "Submission not found") from exc
                except SubmissionRequestError as exc:
                    raise JsonHandlerError(400, "Bad submission_uuid provided") from exc
                except (SubmissionInternalError, Exception) as exc:
                    raise JsonHandlerError(500, "Internal error getting submission info") from exc
            return handler(self, submission_uuid, data, suffix=suffix)
        return wrapped_handler
    return decorator


class StaffGraderMixin:
    """
    Actions to interact with submission locks, blocking other staff from grading assignments while
    grading is in progress.
    """

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    @require_submission_uuid(validate=False)
    def check_submission_lock(self, submission_uuid, data, suffix=""):  # pylint: disable=unused-argument
        """
        Get info about a submission lock. Does not verify that the ID is a valid submission.

        Returns:
        - Serialized submission lock info.
        """
        anonymous_user_id = self.get_anonymous_user_id_from_xmodule_runtime()
        context = {'user_id': anonymous_user_id}

        submission_lock = SubmissionGradingLock.get_submission_lock(submission_uuid) or {}
        return SubmissionLockSerializer(submission_lock, context=context).data

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    @require_submission_uuid(validate=True)
    def claim_submission_lock(self, submission_uuid, data, suffix=''):  # pylint: disable=unused-argument
        """
        Attempt to claim or reclaim a submission lock

        Returns:
        - Serialized submission lock info.

        Raises:
        - 403 in the case of a contested lock
        """
        anonymous_user_id = self.get_anonymous_user_id_from_xmodule_runtime()
        context = {'user_id': anonymous_user_id}

        try:
            submission_lock = SubmissionGradingLock.claim_submission_lock(submission_uuid, anonymous_user_id)
            return SubmissionLockSerializer(submission_lock, context=context).data
        except SubmissionLockContestedError as err:
            raise JsonHandlerError(403, err.get_error_code()) from err

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    @require_submission_uuid(validate=False)
    def delete_submission_lock(self, submission_uuid, data, suffix=''):  # pylint: disable=unused-argument
        """
        Attempt to clear a submission lock.

        Returns:
        - Serialized submission lock info (which in this case would just be {'lock_status': 'unlocked'})

        Raises:
        - 403 in the case of a contested lock
        """
        anonymous_user_id = self.get_anonymous_user_id_from_xmodule_runtime()
        context = {'user_id': anonymous_user_id}

        try:
            SubmissionGradingLock.clear_submission_lock(submission_uuid, anonymous_user_id)
            return SubmissionLockSerializer({}, context=context).data
        except SubmissionLockContestedError as err:
            raise JsonHandlerError(403, err.get_error_code()) from err

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    def batch_delete_submission_lock(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Given a list of submission UUIDs, clear those that we currently have locks for.

        Returns: None, no errors is implicit success

        Raises:
        - 400 in the case of bad params/data
        - 500 for generic errors
        """
        submission_uuids = data.get("submission_uuids")
        if not isinstance(submission_uuids, list):
            raise JsonHandlerError(400, "Body must contain a submission_uuids list")

        anonymous_user_id = self.get_anonymous_user_id_from_xmodule_runtime()
        if not anonymous_user_id:
            raise JsonHandlerError(500, "Failed to get anonymous user ID")

        try:
            SubmissionGradingLock.batch_clear_submission_locks(
                submission_uuids, anonymous_user_id
            )
        except Exception as err:
            raise JsonHandlerError(500, str(err)) from err

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    def list_staff_workflows(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Returns data for the base "list" view, showing a summary of all graded / gradable items in the given assignment

        Example Data Shape:
        {
            submission_uuid:
        }
        """
        # Calculate this once so we don't have to re-query each time
        is_team_assignment = self.is_team_assignment()

        # Fetch staff workflows, annotated with grading_status and lock_status
        staff_workflows = self._bulk_fetch_annotated_staff_workflows(is_team_assignment=is_team_assignment)

        # Lookup additional info like usernames and assessments and determine serializer type
        serializer = TeamSubmissionListSerializer if is_team_assignment else SubmissionListSerializer
        serializer_context = self._get_list_workflows_serializer_context(
            staff_workflows, is_team_assignment=is_team_assignment
        )

        # Serialize workflows with the context, and return the dict of submissions
        result = {}
        for staff_workflow in staff_workflows:
            try:
                serialized_workflow = serializer(staff_workflow, context=serializer_context).data
                result[staff_workflow.identifying_uuid] = serialized_workflow
            except MissingContextException as e:
                log.exception("Failed to serialize workflow %d: %s", staff_workflow.id, str(e), exc_info=True)
        return result

    def _get_list_workflows_serializer_context(self, staff_workflows, is_team_assignment=False):
        """
        Fetch additional required data and models to serialize the response
        """
        # Pull out sets from the workflows for use later
        submission_uuids, workflow_scorer_ids = set(), set()
        for workflow in staff_workflows:
            submission_uuids.add(workflow.identifying_uuid)
            if workflow.scorer_id:
                workflow_scorer_ids.add(workflow.scorer_id)
        course_id = self.get_student_item_dict()['course_id']

        # Fetch user identifier mappings
        if is_team_assignment:
            # Look up the team IDs for submissions so we can later map to team names
            team_submission_uuid_to_team_id = get_team_ids_by_team_submission_uuid(submission_uuids)

            # Look up names for teams
            topic_id = self.selected_teamset_id
            team_id_to_team_name = self.teams_service.get_team_names(course_id, topic_id)

            # Do bulk lookup for scorer anonymous ids (submitting team name is a separate lookup)
            anonymous_id_to_username = map_anonymized_ids_to_usernames(set(workflow_scorer_ids))

            context = {
                'team_submission_uuid_to_team_id': team_submission_uuid_to_team_id,
                'team_id_to_team_name': team_id_to_team_name,
            }
        else:
            # When we look up usernames we want to include all connected learner student ids
            submission_uuid_to_student_id = get_student_ids_by_submission_uuid(
                course_id,
                submission_uuids,
            )

            # Do bulk lookup for all anonymous ids (submitters and scoreres). This is used for the
            # `gradedBy` and `username` fields
            anonymous_id_to_username = map_anonymized_ids_to_usernames(
                set(submission_uuid_to_student_id.values()) | workflow_scorer_ids
            )

            context = {
                'submission_uuid_to_student_id': submission_uuid_to_student_id,
            }

        # Do a bulk fetch of the assessments linked to the workflows, including all connected
        # Rubric, Criteria, and Option models
        submission_uuid_to_assessment = self.bulk_deep_fetch_assessments(staff_workflows)

        context.update({
            'anonymous_id_to_username': anonymous_id_to_username,
            'submission_uuid_to_assessment': submission_uuid_to_assessment,
        })

        return context

    def _bulk_fetch_annotated_staff_workflows(self, is_team_assignment=False):
        """
        Returns: QuerySet of StaffWorkflows, filtered by the current course and item, with the following annotations:
         - current_lock_user: The "owner_id" of the most recent active (created less than TIME_LIMIT ago) lock
         - grading_status: one of
                              * "graded"   - the StaffWorkflow has an associated Assessment
                              * "ungraded" - the StaffWorkflow has no asociated Assessment
        - lock_status: one of
                              * "in-progress" - current_lock_user is the current user's anonymous id.
                                                The current user has an active lock on this submission.
                              * "locked"      - current_lock_user is non-null and not the current user's anonymous id.
                                                Another user has an active lock on this submission.
                              * "unlocked"    - current_lock_user is null
                                                There is no active lock on this submission.
        """
        # Create an unevaluated QuerySet of "active" SubmissionLock objects that refer to the same submission as the
        # "current" workflow
        student_item_dict = self.get_student_item_dict()

        # Return TeamStaffWorkflows for teams, StaffWorkflows for individual
        if is_team_assignment:
            workflow_type = TeamStaffWorkflow
            identifying_uuid = 'team_submission_uuid'
        else:
            workflow_type = StaffWorkflow
            identifying_uuid = 'submission_uuid'

        newest_lock = SubmissionGradingLock.currently_active().filter(
            submission_uuid=OuterRef(identifying_uuid)
        ).order_by(
            '-created_at'
        )

        staff_workflows = workflow_type.objects.filter(
            course_id=student_item_dict['course_id'],
            item_id=student_item_dict['item_id'],
            cancelled_at=None,
        ).annotate(
            current_lock_user=Subquery(newest_lock.values('owner_id')),
        ).annotate(
            grading_status=Case(
                When(assessment__isnull=False, then=Value("graded", output_field=CharField())),
                default=Value("ungraded", output_field=CharField())
            ),
            lock_status=Case(
                When(
                    current_lock_user=student_item_dict['student_id'],
                    then=Value("in-progress", output_field=CharField())
                ),
                When(
                    current_lock_user__isnull=False,
                    then=Value("locked", output_field=CharField())
                ),
                default=Value("unlocked", output_field=CharField())
            )
        )
        return staff_workflows

    def bulk_deep_fetch_assessments(self, staff_workflows):
        """
        Given a list of StaffWorkflows, fetch related Assessments and prefetch
        linked Rubrics, AssessmentParts, Criteria, and Options.

        returns: (dict) mapping identifying uuids to the associated assessment.
        For individual submissions, the key will be the individual Submission uuid, and for
        team submissions, the key will be the TeamSubmission uuid.
        If there is no assessment associated with a submission, it is not included in the dict.
        """
        assessment_id_to_identifying_uuid = {}
        for workflow in staff_workflows:
            if workflow.assessment:
                assessment_id_to_identifying_uuid[workflow.assessment] = workflow.identifying_uuid

        assessments = Assessment.objects.filter(
            pk__in=assessment_id_to_identifying_uuid.keys()
        ).prefetch_related(
            Prefetch(
                "parts",
                queryset=AssessmentPart.objects.select_related('criterion', 'option')
            ),
            "rubric__criteria",
            "rubric__criteria__options"

        ).select_related(
            'rubric',
        ).order_by('-scored_at')
        assessments_by_identifying_uuid = {
            # This "cast" to str is required because StaffWorkflow.assessment is a CharField
            assessment_id_to_identifying_uuid[str(assessment.id)]: assessment
            for assessment in assessments
        }
        return assessments_by_identifying_uuid

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    @require_submission_uuid(validate=True)
    def get_submission_info(self, submission_uuid, _, suffix=''):  # pylint: disable=unused-argument
        """
        Return a dict representation of a submission in the form
        {
            'text': <list of strings representing the raw response for each prompt>
            'files': <list of:>
                {
                    'download_url': <file url>
                    'description': <file description>
                    'name': <file name>
                }
        }
        """
        try:
            if self.is_team_assignment():
                submission = get_team_submission(submission_uuid)
            else:
                submission = get_submission(submission_uuid)
            answer = OraSubmissionAnswerFactory.parse_submission_raw_answer(submission.get('answer'))
        except SubmissionError as err:
            raise JsonHandlerError(404, str(err)) from err
        except VersionNotFoundException as err:
            raise JsonHandlerError(500, str(err)) from err

        return {
            'files': [
                SubmissionDetailFileSerilaizer(file_data).data
                for file_data in self.get_download_urls_from_submission(submission)
            ],
            'text': answer.get_text_responses()
        }

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    @require_submission_uuid(validate=True)
    def get_assessment_info(self, submission_uuid, _, suffix=''):  # pylint: disable=unused-argument
        """
        Returns a dict representation of a staff assessment in the form
        {
            'feedback': <submission-level feedback>
            'points_earned': <earned points>
            'points_possible': <maximum possible points>
            'criteria': list of {
                'name': <criterion name>
                'option': <name of selected option> This may be blank.
                          If so, there are no options defined for the given criterion and it is feedback-only
                'feedback': <feedback for criterion>
            }
        }
        """
        student_item_dict = self.get_student_item_dict()
        course_id = student_item_dict['course_id']
        item_id = student_item_dict['item_id']

        try:
            if self.is_team_assignment():
                workflow = TeamStaffWorkflow.get_team_staff_workflow(course_id, item_id, submission_uuid)
            else:
                workflow = StaffWorkflow.get_staff_workflow(course_id, item_id, submission_uuid)
        except StaffWorkflow.DoesNotExist as ex:
            msg = f"No gradeable submission found with uuid={submission_uuid} in course={course_id} item={item_id}"
            raise JsonHandlerError(404, msg) from ex

        if not workflow.assessment:
            return {}

        assessments = self.bulk_deep_fetch_assessments([workflow])
        if len(assessments) != 1:
            log.error(
                (
                    "[%s] Error looking up assessments. Submission UUID = %s, "
                    "Staff Workflow Id = %s, Staff Workflow Assessment = %s, Assessments = %s"
                ),
                item_id, submission_uuid, workflow.id, workflow.assessment, assessments
            )
            raise JsonHandlerError(500, "Error looking up assessments")

        _, assessment = assessments.popitem()
        return AssessmentSerializer(assessment).data

    @XBlock.json_handler
    @require_course_staff("STUDENT_GRADE")
    @require_submission_uuid(validate=True)
    def submit_staff_assessment(self, submission_uuid, data, suffix=''):  # pylint: disable=unused-argument
        """
        Staff grader-specific wrapper over staff assessments to better handle individual vs team assignments

        data: { (grade data)
            'options_selected': {
                '<criterion_name_1>': <selected_option_name>,
                '<criterion_name_2>': <selected_option_name>,
            },
            'criterion_feedback': {
                '<criterion_name_1>': (string)
            },
            'overall_feedback': (string)
            'submission_uuid': (string)
            'assess_type': (string) one of ['regrade', full-grade']
        }

        Returns: {
            'success': True/False - whether or not the grade submit succeeded
            'msg': String/Empty - error string, if failure occurred
        }
        """
        if self.is_team_assignment():
            success, err_msg = self.do_team_staff_assessment(data, team_submission_uuid=submission_uuid)
        else:
            success, err_msg = self.do_staff_assessment(data)

        return {'success': success, 'msg': err_msg}



""" A mixin for self Assessments. """

import logging

from webob import Response

from openassessment.xblock.utils.user_data import get_user_preferences

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

template_paths = {
    "unavailable": "openassessmentblock/self/oa_self_unavailable.html",
    "cancelled": "openassessmentblock/self/oa_self_cancelled.html",
    "complete": "openassessmentblock/self/oa_self_complete.html",
    "closed": "openassessmentblock/self/oa_self_closed.html",
    "assessment": "openassessmentblock/self/oa_self_assessment.html",
}


def render_self_assessment(api_data):
    if "self-assessment" not in api_data.config_data.assessment_steps:
        return Response("")

    submission_uuid = api_data.submission_data.submission_uuid
    translate = api_data.config_data.translate
    try:
        path, context = self_path_and_context(api_data)
    except Exception:  # pylint: disable=broad-except
        msg = f"Could not retrieve self assessment for submission {submission_uuid}"
        logger.exception(msg)
        return api_data.config_data.render_error(translate("An unexpected error occurred."))
    else:
        return api_data.config_data.render_assessment(path, context)


def self_context(api_data, with_sub=False):
    config_data = api_data.config_data
    user_preferences = get_user_preferences(config_data.user_service)
    context = {
        "allow_multiple_files": config_data.allow_multiple_files,
        "allow_latex": config_data.allow_latex,
        "prompts_type": config_data.prompts_type,
        "xblock_id": config_data.get_xblock_id(),
        "user_timezone": user_preferences["user_timezone"],
        "user_language": user_preferences["user_language"],
    }

    step_data = api_data.self_assessment_data
    # We display the due date whether the problem is open or closed.
    # If no date is set, it defaults to the distant future, in which
    # case we don't display the date.
    if step_data.is_due:
        context["self_due"] = step_data.due_date

    if step_data.is_not_available_yet:
        context["self_start"] = step_data.start_date

    if with_sub and step_data.submission:
        context["rubric_criteria"] = config_data.rubric_criteria_with_labels
        context["self_submission"] = step_data.submission_dict
        if config_data.rubric_feedback_prompt is not None:
            context["rubric_feedback_prompt"] = config_data.rubric_feedback_prompt

        if config_data.rubric_feedback_default_text is not None:
            context["rubric_feedback_default_text"] = config_data.rubric_feedback_default_text

        # Determine if file upload is supported for this XBlock and what kind of files can be uploaded.
        context["file_upload_type"] = config_data.file_upload_type
        context["self_file_urls"] = step_data.file_urls
    return context


def self_path_and_context(api_data):
    """
    Determine the template path and context to use when rendering the self-assessment step.

    Returns:
        tuple of `(path, context)`, where `path` (str) is the path to the template,
        and `context` (dict) is the template context.

    Raises:
        SubmissionError: Error occurred while retrieving the current submission.
        SelfAssessmentRequestError: Error occurred while checking if we had a self-assessment.
    """
    # Import is placed here to avoid model import at project startup.
    step_data = api_data.self_assessment_data

    def path_and_context(path_key, with_sub=False):
        return (template_paths[path_key], self_context(api_data, with_sub))

    if step_data.is_cancelled:
        return path_and_context("cancelled")
    elif step_data.is_self_complete:
        return path_and_context("complete")
    elif step_data.is_self_active or step_data.problem_closed:
        if step_data.assessment is not None:
            return path_and_context("complete")
        elif step_data.problem_closed:
            if step_data.is_not_available_yet:
                return path_and_context("unavailable")
            elif step_data.is_past_due:
                return path_and_context("closed")
        else:
            return path_and_context("assessment", with_sub=True)
    return path_and_context("unavailable", with_sub=True)



""" Views for filesystem backend. """


import hashlib
import json
import os

from django.conf import settings
from django.shortcuts import Http404, HttpResponse
from django.utils import timezone
from django.views.decorators.http import require_http_methods

from . import exceptions
from .backends.base import Settings
from .backends.filesystem import is_download_url_available, is_upload_url_available


@require_http_methods(["PUT", "GET"])
def filesystem_storage(request, key):
    """
    Uploading and download files to the local filesystem backend.
    """
    if request.method == "PUT":
        if not is_upload_url_available(key):
            raise Http404()
        content, metadata = get_content_metadata(request)
        save_to_file(key, content, metadata)
        return HttpResponse()
    elif request.method == "GET":
        if not is_download_url_available(key):
            raise Http404()
        return download_file(key)
    return None


def download_file(key):
    """Returns an HttpResponse to download the corresponding file"""

    file_path = get_file_path(key)
    metadata_path = get_metadata_path(key)
    if not os.path.exists(file_path):
        raise Http404()
    with open(metadata_path) as f:
        metadata = json.load(f)
        content_type = metadata.get("Content-Type", 'application/octet-stream')
    with open(file_path, 'rb') as f:
        response = HttpResponse(f.read(), content_type=content_type)

    file_name = os.path.basename(os.path.dirname(file_path))
    file_extension = Settings.guess_extension(content_type)
    if not file_name.endswith(file_extension):
        file_name += file_extension
    response['Content-Disposition'] = 'attachment; filename=' + file_name
    return response


def get_content_metadata(request):
    """
    Read the content and metadata associated to an HttpRequest.

    Returns:
        request body (str)
        request metadata (dict)
    """

    metadata = {
        "Content-Type": request.META["CONTENT_TYPE"],
        "Date": str(timezone.now()),
        "Content-MD5": hashlib.md5(request.body).hexdigest(),
        "Content-Length": request.META["CONTENT_LENGTH"],
    }
    return request.body, metadata


def save_to_file(key, content, metadata=None):
    """
    Save the content and metadata to a local file determined by the given key.

    Arguments:
        key (str): unique file identifier
        content (str): uploaded file content
        metadata (dict): json-dumpable data
    """
    file_path = get_file_path(key)
    metadata_path = get_metadata_path(key)
    if metadata is None:
        metadata = {}

    safe_save(file_path, content)
    try:
        safe_save(metadata_path, json.dumps(metadata))
    except Exception:
        safe_remove(file_path)
        safe_remove(metadata_path)
        raise


def safe_save(path, content):
    """
    Save content to path. Creates the appropriate directories, if required.

    Raises:
        FileUploadInternalError if the root directory does not exist or if we
        try to save in an unauthorized directory.
    """
    dir_path = os.path.abspath(os.path.dirname(path))
    if not dir_path.startswith(get_bucket_path()):
        raise exceptions.FileUploadRequestError("Uploaded file name not allowed: '%s'" % path)
    root_directory = get_root_directory_path()
    if not os.path.exists(root_directory):
        raise exceptions.FileUploadInternalError("File upload root directory does not exist: %s" % root_directory)
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    mode = "w"
    if isinstance(content, bytes):
        mode = "wb"
    with open(path, mode) as f:
        f.write(content)


def safe_remove(path):
    """Remove a file if it exists.

    Note that an exception will be raised if the file is not writable.
    """
    if os.path.exists(path):
        os.remove(path)
        return True
    return False


def get_file_path(key):
    """
    Returns the path to the content file.
    """
    return os.path.join(get_data_path(key), "content")


def get_metadata_path(key):
    """
    Returns the path to the metadata file.
    """
    return os.path.join(get_data_path(key), "metadata.json")


def get_data_path(key):
    """
    Returns the path to the directory which will store the content and metadata
    files.
    """
    subdirectory = key.replace("..", "").strip("/ ")
    return os.path.join(get_bucket_path(), subdirectory)


def get_bucket_path():
    """
    Returns the path to the bucket directory.
    """
    dir_path = os.path.join(
        get_root_directory_path(),
        Settings.get_bucket_name(),
    )
    return os.path.abspath(dir_path)


def get_root_directory_path():
    """
    Returns the path to the root directory in which bucket directories are stored.

    Raises:
        FileUploadInternalError if the root directory setting does not exist.
    """
    root_dir = getattr(settings, "ORA2_FILEUPLOAD_ROOT", None)
    if not root_dir:
        raise exceptions.FileUploadInternalError("Undefined file upload root directory setting")
    return root_dir



# Generated by Django 2.2.24 on 2021-11-17 21:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('assessment', '0006_TeamWorkflows'),
    ]

    operations = [
        migrations.AlterField(
            model_name='staffworkflow',
            name='assessment',
            field=models.CharField(blank=True, db_index=True, max_length=128, null=True),
        ),
        migrations.AlterField(
            model_name='staffworkflow',
            name='cancelled_at',
            field=models.DateTimeField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='staffworkflow',
            name='grading_completed_at',
            field=models.DateTimeField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='staffworkflow',
            name='grading_started_at',
            field=models.DateTimeField(blank=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name='staffworkflow',
            name='scorer_id',
            field=models.CharField(blank=True, db_index=True, max_length=40),
        ),
    ]



"""
Errors for the staff assessment api.
"""

from .base import AssessmentError


class StaffAssessmentError(AssessmentError):
    """Generic Staff Assessment Error

    Raised when an error occurs while processing a request related to
    staff assessment.

    """


class StaffAssessmentRequestError(StaffAssessmentError):
    """Error indicating insufficient or incorrect parameters in the request.

    Raised when the request does not contain enough information, or incorrect
    information which does not allow the request to be processed.

    """


class StaffAssessmentInternalError(StaffAssessmentError):
    """Error indicating an internal problem independent of API use.

    Raised when an internal error has occurred. This should be independent of
    the actions or parameters given to the API.

    """



"""
Student training step in the OpenAssessment XBlock.
"""
import logging

from webob import Response


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

template_paths = {
    "unavailable": "openassessmentblock/student_training/student_training_unavailable.html",
    "cancelled": "openassessmentblock/student_training/student_training_cancelled.html",
    "complete": "openassessmentblock/student_training/student_training_complete.html",
    "closed": "openassessmentblock/student_training/student_training_closed.html",
    "training": "openassessmentblock/student_training/student_training.html",
    "error": "openassessmentblock/student_training/student_training_error.html",
}


def render_student_training(api_data):
    """
    Render the student training step.

    Args:
        data: Not used.

    Keyword Arguments:
        suffix: Not used.

    Returns:
        unicode: HTML content of the grade step

    """
    if "student-training" not in api_data.config_data.assessment_steps:
        return Response("")

    config_data = api_data.config_data
    try:
        path, context = training_path_and_context(api_data)
    except Exception:  # pylint: disable=broad-except
        submission_uuid = api_data.submission_data.submission_uuid
        msg = f"Could not render Learner Training step for submission {submission_uuid}."
        logger.exception(msg)
        return config_data.render_error(config_data.translate("An unexpected error occurred."))
    else:
        return config_data.render_assessment(path, context)


def training_context(api_data):
    """
    Return the template context used to render the student training step.

    Returns:
        context dict.
    """
    step_data = api_data.student_training_data
    config_data = api_data.config_data

    # Retrieve the status of the workflow.
    # If no submissions have been created yet, the status will be None.
    user_preferences = api_data.config_data.user_preferences

    context = {
        "xblock_id": config_data.get_xblock_id(),
        "allow_multiple_files": config_data.allow_multiple_files,
        "allow_latex": config_data.allow_latex,
        "prompts_type": config_data.prompts_type,
        "user_timezone": user_preferences["user_timezone"],
        "user_language": user_preferences["user_language"],
    }

    if not step_data.has_workflow:
        return context

    if step_data.is_cancelled or step_data.is_complete:
        return context

    # If the problem is closed, then do not allow students to access the training step
    if step_data.is_not_available_yet:
        context["training_start"] = step_data.start_date
        return context
    if step_data.is_past_due:
        context["training_due"] = step_data.due_date
        return context

    if not step_data.training_module:
        return context

    if step_data.is_due:
        context["training_due"] = step_data.due_date

    # Report progress in the student training workflow (completed X out of Y)
    context.update(
        {
            "training_num_available": step_data.num_available,
            "training_num_completed": step_data.num_completed,
            "training_num_current": step_data.num_completed + 1,
        }
    )

    # Retrieve the example essay for the student to submit
    # This will contain the essay text, the rubric, and the options the instructor selected.
    example_context = step_data.example_context
    if not example_context["error_message"]:
        context.update(
            {
                "training_essay": example_context["essay_context"],
                "training_rubric": step_data.example_rubric,
            }
        )

    return context


def training_path_and_context(api_data):
    """
    Return the template path and context used to render the student training step.

    Returns:
        tuple of `(path, context)` where `path` is the path to the template and
            `context` is a dict.

    """
    step_data = api_data.student_training_data

    def path_and_context(path_key):
        return template_paths[path_key], training_context(api_data)

    # Retrieve the status of the workflow.
    # If no submissions have been created yet, the status will be None.
    if not step_data.has_workflow:
        return path_and_context("unavailable")

    # If the student has completed the training step, then show that the step is complete.
    # We put this condition first so that if a student has completed the step, it *always*
    # shows as complete.
    # We're assuming here that the training step always precedes the other assessment steps
    # (peer/self) -- we may need to make this more flexible later.
    if step_data.is_cancelled:
        return path_and_context("cancelled")
    if step_data.is_complete:
        return path_and_context("complete")

    # If the problem is closed, then do not allow students to access the training step
    if step_data.is_not_available_yet:
        return path_and_context("unavailable")
    if step_data.is_past_due:
        return path_and_context("closed")

    # If we're on the training step, show the student an example
    # We do this last so we can avoid querying the student training API if possible.
    if not step_data.training_module:
        return path_and_context("unavailable")

    # Retrieve the example essay for the student to submit
    # This will contain the essay text, the rubric, and the options the instructor selected.
    example_context = step_data.example_context
    if example_context["error_message"]:
        logger.error(example_context["error_message"])
        return path_and_context("error")
    else:
        return path_and_context("training")



"""
Functions that largely mirror the workflow functions specified
in `workflow.api`, but specifically for handling team submissions.
"""
import logging

from django.db import DatabaseError
from django.db.models import Count

from openassessment.workflow.errors import (
    AssessmentWorkflowError,
    AssessmentWorkflowInternalError,
    AssessmentWorkflowRequestError,
    AssessmentWorkflowNotFoundError
)
from openassessment.workflow.models import (
    TeamAssessmentWorkflow,
    AssessmentWorkflowCancellation
)
from openassessment.workflow.serializers import (
    TeamAssessmentWorkflowSerializer,
    AssessmentWorkflowCancellationSerializer
)


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def create_workflow(team_submission_uuid):
    """
    A team submission should only be assessible by staff.  Therefore, unlike
    the analogous `create_workflow()` method for individual submissions,
    we don't accept `steps` or `on_init_params` as parameters to this function,
    since those are only used to indicate which assessment steps (e.g. "peer", "self")
    are to be included in the workflow.

    Raises:
        AssessmentWorkflowInternalError on error
    """
    try:
        team_workflow = TeamAssessmentWorkflow.start_workflow(team_submission_uuid)
        logger.info(
            "Started team assessment workflow for team submission UUID %s",
            team_submission_uuid
        )
        return team_workflow
    except Exception as ex:
        err_msg = (
            "An unexpected error occurred while creating "
            "the workflow for team submission UUID {uuid}"
        ).format(uuid=team_submission_uuid)
        logger.exception(err_msg)
        raise AssessmentWorkflowInternalError(err_msg) from ex


def get_workflow_for_submission(team_submission_uuid):
    """
    Pass through to update_from_assessments. Returns team assessment workflow information
    """
    return update_from_assessments(team_submission_uuid)


def update_from_assessments(team_submission_uuid, override_submitter_requirements=False):
    """
        Like `api.update_from_assessments`, but for teams.  We don't need
        an analogous `assessment_requirements` parameter, because team submissions
        are only assessible by staff (where requirements like "must_grade" and
        "must_be_graded_by" are not supported).

        We also don't need an analogous `course_settings` parameter because there are
        currently no course settings that impact staff grading.

        Raises:
            AssessmentWorkflowInternalError on error
        """
    # Get the wokflow for this submission
    team_workflow = _get_workflow_model(team_submission_uuid)

    # Update the workflow status based on the underlying assessments
    try:
        team_workflow.update_from_assessments(override_submitter_requirements)
        logger.info(
            "Updated workflow for team submission UUID %s",
            team_submission_uuid
        )
    except Exception as exc:
        err_msg = "Could not update team assessment workflow: %s"
        logger.exception(err_msg, exc)
        raise AssessmentWorkflowInternalError(err_msg % exc) from exc

    # Return serialized workflow object
    return _serialized_with_details(team_workflow)


def _serialized_with_details(team_workflow):
    data_dict = TeamAssessmentWorkflowSerializer(team_workflow).data
    data_dict['status_details'] = team_workflow.status_details()
    return data_dict


def _get_workflow_model(team_submission_uuid):
    """
    Returns the `TeamAssessmentWorkflow` model associated with the
    given `team_submission_uuid`.

    Raises:
        AssessmentWorkflowRequestError for incorrect arguments
        AssessmentWorkflowNotFoundError when workflow not found
        AssessmentWorkflowInternalError on error
    """
    if not isinstance(team_submission_uuid, str):
        raise AssessmentWorkflowRequestError("team_submission_uuid must be a string")

    try:
        team_workflow = TeamAssessmentWorkflow.get_by_team_submission_uuid(team_submission_uuid)
    except Exception as exc:
        err_msg = (
            "Could not get team assessment workflow with team_submission_uuid {uuid} due to error: {exc}"
        ).format(uuid=team_submission_uuid, exc=exc)
        logger.exception(err_msg)
        raise AssessmentWorkflowInternalError(err_msg) from exc

    if team_workflow is None:
        err_msg = (
            "No team assessment workflow matching team_submission_uuid {uuid}"
        ).format(uuid=team_submission_uuid)
        raise AssessmentWorkflowNotFoundError(err_msg)

    return team_workflow


def get_status_counts(course_id, item_id):
    """
    Count how many team workflows have each status, for a given item in a course.
    "staff" is the only allowed step for team submissions, so we don't
    need a `steps` parameter here.

    Keyword Arguments:
        course_id (unicode): The ID of the course.
        item_id (unicode): The ID of the item in the course.

    Returns:
        list of dictionaries with keys "status" (str) and "count" (int)

    Example usage:
        >>> get_status_counts("course-v1:edX+DemoX+Demo_Course", "peer-assessment-problem")
        [
            {"status": "staff", "count": 5},
            {"status": "waiting", "count": 43},
            {"status": "done", "count": 0},
        ]
    """
    statuses = TeamAssessmentWorkflow.STEPS + TeamAssessmentWorkflow.STATUSES

    # Remove AI status, valid for workflow logic but not a valid team/student step
    if 'ai' in statuses:
        statuses.remove('ai')

    queryset = TeamAssessmentWorkflow.objects.filter(
        status__in=statuses,
        course_id=course_id,
        item_id=item_id,
    ).values('status').annotate(count=Count('status')).order_by('-created')

    counts_by_status = {status: 0 for status in statuses}

    for row in queryset:
        counts_by_status[row['status']] = row['count']

    return [
        {'status': status, 'count': count}
        for status, count
        in counts_by_status.items()
    ]


def cancel_workflow(team_submission_uuid, comments, cancelled_by_id):
    """
    Add an entry in AssessmentWorkflowCancellation table for a TeamAssessmentWorkflow.

    An TeamAssessmentWorkflow which has been cancelled is no longer included in the
    staff grading pool.

    Team workflows follow the same cancellation workflow,
    but operate on the reference submission.
    """
    try:
        submission_uuid = _get_workflow_model(team_submission_uuid).submission_uuid
        TeamAssessmentWorkflow.cancel_workflow(
            submission_uuid,
            comments,
            cancelled_by_id,
            TeamAssessmentWorkflow.REQUIREMENTS,
            {}
        )
    except Exception as exc:
        err_msg = (
            "Could not cancel team assessment workflow with team_submission_uuid {uuid} due to error: {exc}"
        ).format(uuid=team_submission_uuid, exc=exc)
        logger.exception(err_msg)
        raise AssessmentWorkflowInternalError(err_msg) from exc


def get_assessment_workflow_cancellation(team_submission_uuid):
    """
    Get cancellation information for a team assessment workflow.
    """
    try:
        workflow = _get_workflow_model(team_submission_uuid)
        workflow_cancellation = AssessmentWorkflowCancellation.get_latest_workflow_cancellation(
            workflow.submission_uuid
        )
        return AssessmentWorkflowCancellationSerializer(workflow_cancellation).data if workflow_cancellation else None
    except DatabaseError as ex:
        error_message = (
            "Error finding team assessment workflow cancellation for team submission UUID {uuid}."
        ).format(uuid=team_submission_uuid)
        logger.exception(error_message)
        raise AssessmentWorkflowInternalError(error_message) from ex
    except Exception as exc:
        err_msg = (
            "Could not get workflow cancellation with team_submission_uuid {uuid} due to error: {exc}"
        ).format(uuid=team_submission_uuid, exc=exc)
        logger.exception(err_msg)
        raise AssessmentWorkflowInternalError(err_msg) from exc


def is_workflow_cancelled(team_submission_uuid):
    """
    Check if the team assessment workflow is cancelled
    """
    try:
        workflow = TeamAssessmentWorkflow.get_by_team_submission_uuid(team_submission_uuid)
        return workflow.is_cancelled if workflow else False
    except AssessmentWorkflowError:
        return False



"""
API to encapsulate file CRUD behavior in submissions
"""
import logging

from openassessment.fileupload import api as file_upload_api
from openassessment.fileupload.exceptions import FileUploadError

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class FileAPI:
    def __init__(self, block, team_id):
        self._block = block
        self.team_id = team_id
        self._config_data = block.api_data.config_data

    # Config

    @property
    def max_allowed_uploads(self):
        return self._block.MAX_FILES_COUNT

    @property
    def file_upload_type(self):
        return self._block.file_upload_type

    @property
    def file_manager(self):
        return self._block.file_manager

    # File Uploads
    @property
    def uploaded_files(self):
        """
        Get files uploaded by users, where file uploads are enabled.

        Returns:
        * file_urls (List of File descriptors): files uploaded by user,
          can be empty
        * team_file_urls (List of File descriptors): files uploaded by team,
          can be empty.
        * None when file uploads not enabled.
        """
        if self.file_upload_type:
            file_urls = self.file_manager.file_descriptors(team_id=self.team_id, include_deleted=True)
            team_file_urls = self.file_manager.team_file_descriptors(team_id=self.team_id)
            return {"file_urls": file_urls, "team_file_urls": team_file_urls}
        return None

    # TODO - Determine if we can combine this and uploaded_files
    def get_uploads_for_submission(self):
        """Get a list of uploads for a submission"""
        uploads = self.file_manager.get_uploads(team_id=self.team_id)

        if self._block.is_team_assignment():
            uploads += self.file_manager.get_team_uploads(team_id=self.team_id)

        return uploads

    @property
    def saved_files_descriptions(self):
        return self._config_data.saved_files_descriptions

    # Utils / Actions

    def can_delete_file(self, file_index):
        """
        Is a user allowed to delete this file?

        i.e. are they the uploader and still a member of the team that
        uploaded the file?
        """
        team_id = self.team_id
        key = self.get_file_key(file_index)
        current_user_id = self._block.get_student_item_dict()["student_id"]
        teams_enabled = self._block.is_team_assignment()
        return file_upload_api.can_delete_file(current_user_id, teams_enabled, key, team_id)

    def is_supported_upload_type(self, file_ext, content_type):
        """
        Determine if the uploaded file type/extension is allowed for the
        configured file upload configuration.

        Returns:
            True/False if file type is supported/unsupported
        """
        if self.file_upload_type == "image" and content_type not in self._block.ALLOWED_IMAGE_MIME_TYPES:
            return False

        elif self.file_upload_type == "pdf-and-image" and content_type not in self._block.ALLOWED_FILE_MIME_TYPES:
            return False

        elif self.file_upload_type == "custom" and file_ext.lower() not in self._block.white_listed_file_types:
            return False

        elif file_ext in self._block.FILE_EXT_BLACK_LIST:
            return False

        return True

    def get_file_key(self, file_number):
        """Get file key for {file_number} for current student"""
        student_item_dict = self._block.get_student_item_dict()
        return file_upload_api.get_student_file_key(student_item_dict, index=file_number)

    def get_upload_url(self, key, content_type):
        """Returns key, potentially signed, to upload a file to the file backend"""
        return file_upload_api.get_upload_url(key, content_type)

    def get_download_url(self, file_number):
        """
        Get download URL for a given file number

        Returns
        * URL (string)

        Raises:
        * FileUploadError: when failing to get a download URL
        """
        file_key = self.get_file_key(file_number)
        url = ""
        try:
            if file_key:
                url = file_upload_api.get_download_url(file_key)
        except FileUploadError as exc:
            logger.exception(
                "FileUploadError: Download url for file key %s failed with error %s",
                file_key,
                exc,
                exc_info=True,
            )

        return url

    def delete_uploaded_file(self, file_index):
        self.file_manager.delete_upload(file_index)

    def get_download_urls_from_submission(self, submission):
        return self._block.get_download_urls_from_submission(submission)

    def get_files_info_from_user_state(self, username):
        return self._block.get_files_info_from_user_state(username)

    def get_all_upload_urls_for_user(self, username_or_email):
        return self._block.get_all_upload_urls_for_user(username_or_email)

    def get_allowed_file_types_or_preset(self):
        return self._block.get_allowed_file_types_or_preset



""" Admin file of workflow app """


from django.contrib import admin

from .models import AssessmentWorkflow, AssessmentWorkflowStep, TeamAssessmentWorkflow


class AssessmentWorkflowStepInline(admin.StackedInline):
    """Assessment Workflow Step"""
    model = AssessmentWorkflowStep
    extra = 0


class AssessmentWorkflowAdmin(admin.ModelAdmin):
    """Admin for the user's overall workflow through open assessment.

    Unlike many of the other models, we allow editing here. This is so that we
    can manually move a user's entry to "done" and give them a separate score
    in the submissions app if that's required. Unlike rubrics and assessments,
    there is no expectation of immutability for `AssessmentWorkflow`.
    """
    list_display = (
        'status', 'submission_uuid', 'course_id', 'item_id', 'status_changed'
    )
    list_filter = ('status',)
    search_fields = ('submission_uuid', 'course_id', 'item_id')
    inlines = (AssessmentWorkflowStepInline,)


class TeamAssessmentWorkflowAdmin(admin.ModelAdmin):
    """
    Admin for TeamAssessmentWorkflows.
    """

    list_display = (
        'id', 'status', 'team_submission_uuid', 'submission_uuid', 'course_id', 'item_id', 'status_changed'
    )
    list_filter = ('status',)
    search_fields = ('team_submission_uuid', 'submission_uuid', 'course_id', 'item_id')

    inlines = (AssessmentWorkflowStepInline,)


admin.site.register(AssessmentWorkflow, AssessmentWorkflowAdmin)
admin.site.register(TeamAssessmentWorkflow, TeamAssessmentWorkflowAdmin)



"""
External API for ORA Student Training data
"""
from openassessment.assessment.api.student_training import (
    get_num_completed,
    get_training_example,
)

from openassessment.xblock.utils.data_conversion import (
    convert_training_examples_list_to_dict,
    create_submission_dict,
)
from openassessment.xblock.apis.step_data_api import StepDataAPI


class StudentTrainingAPI(StepDataAPI):
    def __init__(self, block):
        super().__init__(block, "student-training")

    def __repr__(self):
        if self.training_module:
            return "{0}".format(
                {
                    "due_date": self.due_date,
                    "has_workflow": self.has_workflow,
                    "is_cancelled": self.is_cancelled,
                    "is_complete": self.is_complete,
                    "is_due": self.is_due,
                    "is_not_available_yet": self.is_not_available_yet,
                    "is_past_due": self.is_past_due,
                    "num_available": self.num_available,
                    "num_completed": self.num_completed,
                    "start_date": self.start_date,
                    "training_module": self.training_module,
                }
            )
        return "{0}".format(
            {
                "due_date": self.due_date,
                "has_workflow": self.has_workflow,
                "is_cancelled": self.is_cancelled,
                "is_complete": self.is_complete,
                "is_due": self.is_due,
                "is_not_available_yet": self.is_not_available_yet,
                "is_past_due": self.is_past_due,
                "start_date": self.start_date,
                "training_module": self.training_module,
            }
        )

    @property
    def example(self):
        return get_training_example(
            self._block.submission_uuid,
            {
                "prompt": self.config_data.prompt,
                "criteria": self._block.rubric_criteria_with_labels,
            },
            self.examples,
        )

    @property
    def example_context(self):
        context, error_message = self._parse_example(self.example)
        return {"error_message": error_message, "essay_context": context}

    @property
    def example_rubric(self):
        rubric = self.example["rubric"]
        return {
            "criteria": rubric["criteria"],
            "points_possible": rubric["points_possible"],
        }

    @property
    def examples(self):
        return convert_training_examples_list_to_dict(self.training_module["examples"])

    @property
    def has_workflow(self):
        return self.workflow_data.has_status

    @property
    def is_cancelled(self):
        return self.workflow_data.is_cancelled

    @property
    def is_complete(self):
        state = self.workflow_data
        return state.has_status and not (state.is_cancelled or state.is_training)

    @property
    def num_available(self):
        return len(self.training_module["examples"])

    @property
    def num_completed(self):
        return get_num_completed(self._block.submission_uuid)

    @property
    def training_module(self):
        return self.config_data.get_assessment_module("student-training")

    def _parse_answer_dict(self, answer):
        """
        Helper to parse answer as a fully-qualified dict.
        """
        parts = answer.get("parts", [])
        if parts and isinstance(parts[0], dict):
            if isinstance(parts[0].get("text"), str):
                return create_submission_dict({"answer": answer}, self.config_data.prompts)
        return None

    def _parse_answer_list(self, answer):
        """
        Helper to parse answer as a list of strings.
        """
        if answer and isinstance(answer[0], str):
            return self._parse_answer_string(answer[0])
        elif not answer:
            return self._parse_answer_string("")
        return None

    def _parse_answer_string(self, answer):
        """
        Helper to parse answer as a plain string
        """
        return create_submission_dict({"answer": {"parts": [{"text": answer}]}}, self.config_data.prompts)

    def _parse_example(self, example):
        """
        EDUCATOR-1263: examples are serialized in a myriad of different ways, we need to be robust to all of them.

        Types of serialized example['answer'] we handle here:
        -fully specified: {'answer': {'parts': [{'text': <response_string>}]}}
        -list of string: {'answer': [<response_string>]}
        -just a string: {'answer': <response_string>}
        """
        if not example:
            return (
                {},
                "No training example was returned fromt he API for student with Submission UUID {}".format(
                    self._block.submission_uuid
                ),
            )
        answer = example["answer"]
        submission_dict = None
        if isinstance(answer, str):
            submission_dict = self._parse_answer_string(answer)
        elif isinstance(answer, dict):
            submission_dict = self._parse_answer_dict(answer)
        elif isinstance(answer, list):
            submission_dict = self._parse_answer_list(answer)
        return (submission_dict, "") or (
            {},
            f"Improperly formatted example, cannot render student training.  Example: {example}",
        )



"""
Serializers specific to peer assessment.
"""


from rest_framework import serializers

from openassessment.assessment.models import (AssessmentFeedback, AssessmentFeedbackOption, PeerWorkflow,
                                              PeerWorkflowItem)

from .base import AssessmentSerializer


class AssessmentFeedbackOptionSerializer(serializers.ModelSerializer):
    """
    Serialize an `AssessmentFeedbackOption` model.
    """

    class Meta:
        model = AssessmentFeedbackOption
        fields = ('text',)


class AssessmentFeedbackSerializer(serializers.ModelSerializer):
    """
    Serialize feedback in response to an assessment.
    """
    assessments = AssessmentSerializer(many=True, default=None, required=False)
    options = AssessmentFeedbackOptionSerializer(many=True, default=None, required=False)

    class Meta:
        model = AssessmentFeedback
        fields = ('submission_uuid', 'feedback_text', 'assessments', 'options')


class PeerWorkflowSerializer(serializers.ModelSerializer):
    """Representation of the PeerWorkflow.

    A PeerWorkflow should not be exposed to the front end of any question.  This
    model should only be exposed externally for administrative views, in order
    to visualize the Peer Workflow.

    """

    class Meta:
        model = PeerWorkflow
        fields = (
            'student_id',
            'item_id',
            'course_id',
            'submission_uuid',
            'created_at',
            'completed_at'
        )


class PeerWorkflowItemSerializer(serializers.ModelSerializer):
    """Representation of the PeerWorkflowItem

    As with the PeerWorkflow, this should not be exposed to the front end. This
    should only be used to visualize the Peer Workflow in an administrative
    view.

    """

    class Meta:
        model = PeerWorkflowItem
        fields = (
            'scorer',
            'author',
            'submission_uuid',
            'started_at',
            'assessment',
            'scored'
        )



"""
External API for ORA Submission data
"""
from copy import deepcopy
import logging

from submissions.team_api import get_team_submission

from openassessment.xblock.utils.data_conversion import (
    format_files_for_submission,
    prepare_submission_for_serialization,
    update_saved_response_format,
)
from openassessment.xblock.utils.resolve_dates import DISTANT_FUTURE
from openassessment.xblock.apis.step_data_api import StepDataAPI
from openassessment.xblock.apis.submissions.errors import (
    EmptySubmissionError,
    NoTeamToCreateSubmissionForError,
)
from openassessment.xblock.apis.submissions.file_api import FileAPI

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class SubmissionAPI(StepDataAPI):
    def __init__(self, block):
        super().__init__(block, "submission")
        self._workflow_data = block.workflow_data
        self.files = FileAPI(block, self.team_id)

    @property
    def workflow(self):
        return self.workflow_data.workflow

    # Submission Statuses

    @property
    def has_submitted(self):
        return bool(self.workflow)

    @property
    def has_been_cancelled(self):
        return self.workflow and self.workflow["status"] == "cancelled"

    @property
    def cancellation_info(self):
        if self.config_data.is_team_assignment():
            return self.workflow_data.get_team_workflow_cancellation_info(self.team_submission_uuid)
        else:
            return self.workflow_data.get_workflow_cancellation_info(self.submission_uuid)

    @property
    def has_received_final_grade(self):
        return self.workflow and self.workflow["status"] == "done"

    @property
    def peer_step_incomplete(self):
        return "peer" in self.workflow_data.status_details and not self.workflow_data.is_peer_complete

    @property
    def self_step_incomplete(self):
        return "self" in self.workflow_data.status_details and not self.workflow_data.is_self_complete

    # Submission Access information

    @property
    def problem_is_inaccessible(self):
        return super().problem_closed

    @property
    def problem_inaccessible_reason(self):
        return super().closed_reason

    @property
    def problem_is_past_due(self):
        return super().is_past_due

    @property
    def due_date(self):
        """A problem due in the "distant future" has, in effect no due date"""
        if super().due_date < DISTANT_FUTURE:
            return super().due_date
        return None

    @property
    def team_previously_submitted_without_student(self):
        """
        Determine if student has joined a team that already submitted without them.
        They would be blocked from submitting or joining this submission.
        """
        if not self.config_data.teams_enabled:
            return False
        return not self.has_submitted and self._block.does_team_have_submission(self.team_id)

    # Submission / response data

    @property
    def has_saved(self):
        return self._block.has_saved

    @has_saved.setter
    def has_saved(self, value):
        self._block.has_saved = value

    @property
    def saved_response(self):
        """Return a saved response for a student / team when they haven't submitted"""
        return update_saved_response_format(self.config_data.saved_response)

    @saved_response.setter
    def saved_response(self, value):
        self._block.saved_response = value

    @property
    def student_submission(self):
        """Return a saved response or a student / team submission"""
        return self._block.get_user_submission(self.workflow["submission_uuid"])

    @property
    def submission_uuid(self):
        """Return a submission_uuid or None if the user hasn't submitted"""
        return self.workflow.get("submission_uuid")

    # Team Info

    @property
    def team_id(self):
        """
        Return the Team ID for this user / submission.

        This is either:
        1) None - when this is not a team assignment or user is not on a team
        2) The current team ID - when the active team has not yet submitted
        3) A previous team ID - the team ID associated with a previous
           submission, even if the user has since switched to a new team.

        Returns:
        * UUID or None
        """
        if self.is_team_assignment:
            if not self.has_submitted:
                return self.config_data.get_team_info().get("team_id")
            return get_team_submission(self.team_submission_uuid).get("team_id")
        return None

    @property
    def team_submission_uuid(self):
        return self.workflow.get("team_submission_uuid")

    # Submission config

    @property
    def is_individual_assignment(self):
        return not self.is_team_assignment

    @property
    def is_team_assignment(self):
        return self.config_data.is_team_assignment()

    @property
    def response_config(self):
        """
        Context needed to author a response
        """
        response_config = {
            "prompts_type": self.config_data.prompts_type,
            # Response
            "text_response": self.config_data.text_response,
            "text_response_editor": self.config_data.text_response_editor,
            "show_rubric_during_response": self.config_data.show_rubric_during_response,
            "allow_latex": self.config_data.allow_latex,
            # File upload
            "enable_delete_files": False,
            "file_upload_response": self.config_data.file_upload_response,
            "file_upload_type": self.config_data.file_upload_type,
            "allow_multiple_files": self.config_data.allow_multiple_files,
        }

        if self.config_data.file_upload_type:
            response_config["white_listed_file_types"] = [
                "." + ext for ext in self._block.get_allowed_file_types_or_preset()
            ]

        if self.config_data.show_rubric_during_response:
            response_config["rubric_criteria"] = deepcopy(self.config_data.rubric_criteria_with_labels)

        if self.due_date is not None:
            response_config["date_config_type"] = self.config_data.date_config_type

        return response_config

    # Actions

    def submission_is_empty(self, submission_dict):
        """
        Check if student_sub_dict has any submission content so that we don't
        create empty submissions.

        If there are no text responses and no file responses, raise an EmptySubmissionError

        Args:
        * submission_dict

        Returns:
        * Boolean - whether the submission is empty
        """
        has_content = False

        # Does the student_sub_dict have any non-zero-length strings in 'parts'?
        has_content |= any(part.get("text", "") for part in submission_dict.get("parts", []))

        # Are there any file_keys in student_sub_dict?
        has_content |= len(submission_dict.get("file_keys", [])) > 0

        return not has_content

    def create_submission(self, student_item_dict, submission_data):
        """Creates submission for the submitted assessment response or a list for a team assessment."""
        # Import is placed here to avoid model import at project startup.
        from submissions import api

        # Serialize the submission
        submission_dict = prepare_submission_for_serialization(submission_data)

        # Add files
        uploaded_files = self.files.get_uploads_for_submission()
        submission_dict.update(format_files_for_submission(uploaded_files))

        # Validate
        if self.submission_is_empty(submission_dict):
            raise EmptySubmissionError

        # Create submission
        submission = api.create_submission(student_item_dict, submission_dict)
        self.workflow_data.create_workflow(submission["uuid"])

        # Set student submission_uuid
        self._block.submission_uuid = submission["uuid"]

        # Emit analytics event...
        self.config_data.publish_event(
            "openassessmentblock.create_submission",
            {
                "submission_uuid": submission["uuid"],
                "attempt_number": submission["attempt_number"],
                "created_at": submission["created_at"],
                "submitted_at": submission["submitted_at"],
                "answer": submission["answer"],
            },
        )

        return submission

    def create_team_submission(self, student_item_dict, submission_data):
        """A student submitting for a team should generate matching submissions for every member of the team."""

        if not self.config_data.has_team:
            student_id = student_item_dict["student_id"]
            course_id = self.config_data.course_id
            msg = f"Student {student_id} has no team for course {course_id}"
            logger.exception(msg)
            raise NoTeamToCreateSubmissionForError(msg)

        # Import is placed here to avoid model import at project startup.
        from submissions import team_api

        team_info = self.config_data.get_team_info()

        # Serialize the submission
        submission_dict = prepare_submission_for_serialization(submission_data)

        # Add files
        uploaded_files = self.files.get_uploads_for_submission()
        submission_dict.update(format_files_for_submission(uploaded_files))

        # Validate
        if self.submission_is_empty(submission_dict):
            raise EmptySubmissionError

        submitter_anonymous_user_id = self.config_data.get_anonymous_user_id_from_xmodule_runtime()
        user = self.config_data.get_real_user(submitter_anonymous_user_id)

        anonymous_student_ids = self.config_data.get_anonymous_user_ids_for_team()
        submission = team_api.create_submission_for_team(
            self.config_data.course_id,
            student_item_dict["item_id"],
            team_info["team_id"],
            user.id,
            anonymous_student_ids,
            submission_dict,
        )

        self.workflow_data.create_team_workflow(submission["team_submission_uuid"])

        # Emit analytics event...
        self.config_data.publish_event(
            "openassessmentblock.create_team_submission",
            {
                "submission_uuid": submission["team_submission_uuid"],
                "team_id": team_info["team_id"],
                "attempt_number": submission["attempt_number"],
                "created_at": submission["created_at"],
                "submitted_at": submission["submitted_at"],
                "answer": submission["answer"],
            },
        )
        return submission
